"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import networkx as nx
from networkx import DiGraph

from qrisp.circuit import (
    ControlledOperation,
    PTControlledOperation,
    QuantumCircuit,

)
from qrisp.logic_synthesis import LogicSynthGate
from qrisp.uncomputation.type_checker import is_permeable, is_qfree
from qrisp.circuit import fast_append


class UnqompNode:
    def __init__(self, name, instr=None):
        self.name = name
        self.targets = []
        self.controls = []
        self.uncomputed_node = None

        self.instr = instr
        if instr is None:
            self.hash = hash(self.name)
        else:
            self.hash = hash(instr)

    def __hash__(self):
        return self.hash

    def __str__(self):
        if self.instr is None:
            return self.name
        else:
            return str(self.instr)

    def __repr__(self):
        return str(self)
    
    def __eq__(self, other):
        return self.hash == other.hash


def qc_from_dag(dag):
    res_qc = dag.original_qc.clearcopy()
    from qrisp.core.compilation import topological_sort
        
    with fast_append():
        for node in nx.topological_sort(dag):
            if node.instr:
                res_qc.append(node.instr.op, node.instr.qubits, node.instr.clbits)

    return res_qc


def dag_from_qc(qc, remove_init_nodes=False):
    dag = DiGraph()

    recent_node_dic = {}
    node_counter = {}
    init_nodes = {}

    for i in range(len(qc.qubits)):
        node = UnqompNode("qubit_" + str(i) + "_0")

        node.qubit = qc.qubits[i]
        
        # dag.add_node(node)
        dag._succ[node] = {}
        dag._pred[node] = {}
        dag._node[node] = {}
        
        recent_node_dic[qc.qubits[i]] = node
        node_counter[qc.qubits[i]] = 1
        init_nodes[qc.qubits[i]] = node
    
    for i in range(len(qc.clbits)):
        
        node = UnqompNode("clbit_" + str(i) + "_0")

        node.clbit = qc.clbits[i]

        # dag.add_node(node)
        dag._succ[node] = {}
        dag._pred[node] = {}
        dag._node[node] = {}
        
        
        recent_node_dic[qc.clbits[i]] = node
        node_counter[qc.clbits[i]] = 1
        init_nodes[qc.clbits[i]] = node
        

    dag.init_nodes = init_nodes

    dealloc_nodes = []

    for i in range(len(qc.data)):
        instr = qc.data[i]

        # if instr.op.name == "barrier":
        # continue

        # node = UnqompNode(str(qc.qubits.index(instr.qubits[-1])) + "_"
        #   + str(node_counter[instr.qubits[-1]]), instr)
        node = UnqompNode(str(node_counter[instr.qubits[-1]]), instr)

        if instr.op.name == "qb_dealloc":
            dealloc_nodes.append(node)

        # dag.add_node(node)
        dag._succ[node] = {}
        dag._pred[node] = {}
        dag._node[node] = {}
        
        
        for j in range(len(instr.qubits)):
            qb = instr.qubits[j]

            if is_permeable(instr.op, [j]):
                edge_type = "control"
                node.controls.append(qb)
            else:
                edge_type = "target"

            if (
                dag.has_edge(recent_node_dic[qb], node)
                and dag.get_edge_data(recent_node_dic[qb], node)["edge_type"]
                != "anti_dependency"
            ):
                dag.get_edge_data(recent_node_dic[qb], node)["qubits"].append(qb)
            else:
                dag.add_edge(
                    recent_node_dic[qb], node, edge_type=edge_type, qubits=[qb]
                )

            successors = list(dag.successors(recent_node_dic[qb]))
            if len(successors) > 1:
                if edge_type == "target":
                    for s in successors:
                        if s == node:
                            continue
                        elif not dag.has_edge(s, node):
                            if (
                                dag.get_edge_data(recent_node_dic[qb], s)["edge_type"]
                                == "control"
                            ):
                                dag.add_edge(s, node, edge_type="anti_dependency")

            if edge_type == "target":
                recent_node_dic[qb] = node
                node_counter[qb] += 1
                node.targets.append(qb)
        for j in range(len(instr.clbits)):
            cb = instr.clbits[j]
            dag.add_edge(recent_node_dic[cb], node, edge_type="anti_dependency")
            recent_node_dic[cb] = node

    dag.original_qc = qc
    dag.recent_node_dic = recent_node_dic

    if remove_init_nodes:
        # Remove init nodes
        dag.remove_nodes_from(list(init_nodes.values()))

    return dag


def uncompute_node(dag, uncomp_node, uncomp_qbs, recompute_qubits=[]):
    from qrisp.core import topological_sort    

    a_star_n_list = [dag.recent_node_dic[qb] for qb in set(uncomp_node.targets)]
    a_star_n_list = list(set(a_star_n_list))
    
    if not is_qfree(uncomp_node.instr.op):
        raise Exception(f"Tried to uncompute non-qfree instruction {uncomp_node.instr}")

    ctrls = [
        n
        for n in dag.predecessors(uncomp_node)
        if dag.get_edge_data(n, uncomp_node)["edge_type"] == "control"
    ]

    for i in range(len(ctrls)):
        if not ctrls[i].uncomputed_node is None:
            ctrls[i] = ctrls[i].uncomputed_node

    # Replace controlled operations by phase tolerant controlled operations
    if isinstance(uncomp_node.instr.op, ControlledOperation):
        op = uncomp_node.instr.op

        if op.method == "auto" and len(op.controls) < 5 or op.method == "gray":
            if len(op.controls) != 1:
                uncomp_node.instr.op = PTControlledOperation(
                    op.base_operation,
                    num_ctrl_qubits=len(op.controls),
                    ctrl_state=op.ctrl_state,
                    method = "gray_pt"
                )

    # Replace results of logic synthesis by phase tolerant logic synthesis
    if isinstance(uncomp_node.instr.op, LogicSynthGate):
        if uncomp_node.instr.op.logic_synth_method == "gray":
            from qrisp import QuantumVariable

            tt = uncomp_node.instr.op.tt

            input_qv = QuantumVariable(tt.bit_amount)
            output_qv = QuantumVariable(tt.shape[1], qs=input_qv.qs)

            tt.q_synth(input_qv, output_qv, method="gray_pt")

            temp = input_qv.qs.data[-1].op
            temp.name = uncomp_node.instr.op.name
            
            uncomp_node.instr.op = temp

    new_instr = uncomp_node.instr.inverse()

    reversed_node = UnqompNode(uncomp_node.name + "_*", new_instr)
    reversed_node.controls = uncomp_node.controls
    reversed_node.targets = uncomp_node.targets

    uncomp_node.uncomputed_node = reversed_node

    dag.add_node(reversed_node)
    
    for qb in reversed_node.targets:
        dag.recent_node_dic[qb] = reversed_node

    for i in range(len(ctrls)):
        c = ctrls[i]

        control_edge_qubits = dag.get_edge_data(c, uncomp_node)["qubits"]

        if set(control_edge_qubits).intersection(recompute_qubits):
            for k in dag.successors(c):
                if dag.get_edge_data(c, k)["edge_type"] == "target" and set(
                    dag.get_edge_data(c, k)["qubits"]
                ).intersection(control_edge_qubits):
                    if k.uncomputed_node is None:
                        return True

                    ctrls[i] = k.uncomputed_node
                    break
            else:
                return True

        dag.add_edge(
            ctrls[i], reversed_node, edge_type="control", qubits=control_edge_qubits
        )

    for a_star_n in a_star_n_list:
        dag.add_edge(
            a_star_n,
            reversed_node,
            edge_type="target",
            qubits=list(
                set(uncomp_qbs).intersection(a_star_n.targets).intersection(new_instr.qubits)
            ),
        )

        for v in dag.successors(a_star_n):
            if dag.get_edge_data(a_star_n, v)["edge_type"] == "control":
                dag.add_edge(v, reversed_node, edge_type="anti_dependency")

    for c in ctrls:
        for v in dag.successors(c):
            if dag.get_edge_data(c, v)["edge_type"] == "target":
                target_edge_qubits = dag.get_edge_data(c, v)["qubits"]
                control_edge_qubits = dag.get_edge_data(c, uncomp_node)["qubits"]

                if set(target_edge_qubits).intersection(control_edge_qubits):
                    pass
                    dag.add_edge(reversed_node, v, edge_type="anti_dependency")

    return False


def uncompute_qc(qc, uncomp_qbs, recompute_qubits=[]):
    
    with fast_append():
        qc_new = qc.copy()
        previous_instructions = []
        follow_up_instructions = []
    
        while len(qc_new.data):
            instr = qc_new.data[0]
            if set(uncomp_qbs + recompute_qubits).intersection(instr.qubits):
                break
            previous_instructions.append(qc_new.data.pop(0))
            
    
        while len(qc_new.data):
            instr = qc_new.data[-1]
            if set(uncomp_qbs + recompute_qubits).intersection(instr.qubits):
                break
            follow_up_instructions.append(qc_new.data.pop(-1))
    
        follow_up_instructions = follow_up_instructions[::-1]
    
        dag = dag_from_qc(qc_new)
    
        lin = list(nx.topological_sort(dag))
        lin.reverse()
    
        for i in range(len(lin)):
            # for node in lin:
    
            node = lin[i]
    
            if node.instr is None:
                continue
    
            if not set(node.targets).intersection(uncomp_qbs):
                continue
    
            # if node.instr.op.name == "qb_alloc":# , "qb_alloc", "qb_dealloc"]:
            #     continue
    
            # This function checks if there is more than one allocation gate in
            # the upcoming chain of nodes. This prevents instructions which were
            # created due to previous recomputations to be included in further recomputations
            if detect_double_alloc(lin[i:], node.instr.qubits):
                continue
    
            if set(node.targets).issubset(uncomp_qbs) and node.instr and node.targets:
                recompute = uncompute_node(dag, node, uncomp_qbs, recompute_qubits)
                if recompute:
                    return uncompute_qc(
                        qc,
                        uncomp_qbs
                        + list(set(node.controls).intersection(recompute_qubits)),
                        recompute_qubits,
                    )
                continue
    
            off_target_qubits = list(set(node.targets) - set(uncomp_qbs))
            if off_target_qubits:
                not_uncomputable_qubits = []
                for qb in off_target_qubits:
                    if qb.allocated:
                        not_uncomputable_qubits.append(qb)
    
                if not_uncomputable_qubits:
                    raise Exception(
                        f'Uncomputation failed because gate "{node.instr.op.name}" needs to be uncomputed '
                        f'but is also targeting qubits {not_uncomputable_qubits} which are not up for uncomputation'
                    )
                else:
                    return uncompute_qc(
                        qc,
                        uncomp_qbs + off_target_qubits,
                        recompute_qubits=recompute_qubits,
                    )
    
        try:
            uncomputed_qc = qc_from_dag(dag)
        except nx.NetworkXUnfeasible:
            raise Exception("Cyclic dependency detected in DAG during uncomputation")
    
        uncomputed_qc.data = (
            previous_instructions + uncomputed_qc.data + follow_up_instructions
        )
        
    
        return uncomputed_qc


def detect_double_alloc(lin, qubits):
    qb_allocs = []
    for node in lin:
        if not node.instr:
            continue
        if node.instr.op.name == "qb_alloc" and node.instr.qubits[0] in qubits:
            if node.instr.qubits[0] in qb_allocs:
                return True
            else:
                qb_allocs.append(node.instr.qubits[0])

    return False



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# -*- coding: utf-8 -*-

# This file contains some functions that enhance the speed
# of SparseBiArray manipulation

import numpy as np
from numba import int64, int32, njit, prange, vectorize
from qrisp.simulator.numerics_config import float_tresh, cutoff_ratio
from scipy.sparse import coo_array

# It can happen that the coo-matrix multiplication puts two data entries
# for the same index. These function recognizes and sums these duplicates
def sum_duplicates(nz_indices, data):
    new_indices, reconstr_indices = np.unique(nz_indices, return_inverse=True)

    if len(nz_indices) > 1e4:
        new_data = sum_duplicates_aux_jitted(new_indices, data, reconstr_indices)
    else:
        new_data = sum_duplicates_aux(new_indices, data, reconstr_indices)

    return new_indices, new_data


def sum_duplicates_aux(new_indices, old_data, reconstr_indices):
    new_data = np.zeros(new_indices.shape, dtype=old_data.dtype)

    for i in range(len(reconstr_indices)):
        new_data[reconstr_indices[i]] += old_data[i]

    return new_data


@njit(nogil=True, cache=True)
def sum_duplicates_aux_jitted(new_indices, old_data, reconstr_indices):
    new_data = np.zeros(new_indices.shape, dtype=old_data.dtype)

    for i in range(len(reconstr_indices)):
        new_data[reconstr_indices[i]] += old_data[i]

    return new_data


# This function rounds entries, which are below a certain float thresh and removes the
# tracked indices.
# @njit(parallel = True)
def elim_zeros(nz_indices, data, decimals=8):
    # return nz_indices, data
    # data = np.around(data, decimals=decimals, out=data)

    # mask = np.abs(data) > 10**-decimals
    mask = np.nonzero(data)[0]

    return nz_indices[mask], data[mask]


# This function applies an index permutation onto
# flattened indices.
# Consider the situation that we have the two flattened indices
# [1,3]
# of a 2x2 matrix.
# If we now want to permute the axes, we permute the bit level
# representation of these numbers
# ["01", "10"] ==> ["10", "01"]


@njit(nogil=True, cache=True)
def permute_axes(index, permutation):
    permutation = -permutation + len(permutation) - 1
    permutation = permutation[::-1]
    new_index = np.zeros(index.size, dtype=index.dtype)
    for i in range(len(permutation)):
        if permutation[i] == i:
            # new_index = new_index | (index&(1<<permutation[i]))
            for j in prange(len(new_index)):
                new_index[j] = new_index[j] | (index[j] & (1 << permutation[i]))
        else:
            # new_index = new_index |
            # (((index&(1<<permutation[i]))>>permutation[i])<<(i))
            for j in prange(len(new_index)):
                new_index[j] = new_index[j] | (
                    ((index[j] & (1 << permutation[i])) >> permutation[i]) << (i)
                )

    return new_index


# This function retrieves the coordinates from
# flattened coordinades for a given shape of length two
def get_coordinates(indices, shape):
    if shape[0] < 1<<31:
        coordinates_0 = get_coordinates_0_32(indices, int(np.log2(shape[1])))
    else:
        coordinates_0 = get_coordinates_0_64(indices, int(np.log2(shape[1])))
        
    if shape[1] < 1<<31:
        coordinates_1 = get_coordinates_1_32(indices, int(np.log2(shape[1])))
    else:
        coordinates_1 = get_coordinates_1_64(indices, int(np.log2(shape[1])))

    return coordinates_0, coordinates_1
        


@vectorize([int64(int64, int64)])
def get_coordinates_0_64(index, bit_shape_1):
    return index >> bit_shape_1


@vectorize([int64(int64, int64)])
def get_coordinates_1_64(index, bit_shape_0):
    return index & ((1 << bit_shape_0) - 1)

@vectorize([int32(int64, int64)])
def get_coordinates_0_32(index, bit_shape_1):
    return index >> bit_shape_1

@vectorize([int32(int64, int64)])
def get_coordinates_1_32(index, bit_shape_0):
    return index & ((1 << bit_shape_0) - 1)



# This function basically inverts get_coordinates.
# Ie. it generates flat coordinates from row and column indices
def gen_flat_coords(col, row, shape):
    return gen_flat_coords_vec(col, row, int(np.log2(shape[1])))
    
@vectorize([int64(int64, int64, int64)], cache=True)
def gen_flat_coords_vec(col, row, bit_shape_1):
    return row << bit_shape_1 | col



# This function constructs a flattened numpy array from
# given flattened coordinates and data
@njit(nogil=True, cache=True)
def construct_flat_array(flat_coords, data, size):
    res = np.zeros(size, dtype=data.dtype)

    for i in range(len(flat_coords)):
        res[flat_coords[i]] += data[i]

    return res

def construct_flat_array_no_jit(flat_coords, data, size):
    res = np.zeros(size, dtype=data.dtype)

    for i in range(len(flat_coords)):
        res[flat_coords[i]] += data[i]

    return res


@njit(cache=True)
def extract_bit_range(index, start, stop):
    breadth = stop - start
    return (index & (((1 << breadth) - 1) << start)) >> start


@njit(cache=True)
def insert_bit_range(target, bit_string, start):
    target |= bit_string << start


@njit(nogil=True, cache=True)
def jitted_permuter(index, new_index, bit_partition, permuted_bit_partition, perm):
    # for i in range(len(perm)):
    #     # bit_range = extract_bit_range(index, bit_partition[i], bit_partition[i+1])
    #     # insert_bit_range(new_index, bit_range, permuted_bit_partition[perm[i]])
    #     #Inlined version more efficient:
    #     new_index |= (((index&((((1<<(bit_partition[i+1] -
    #     bit_partition[i]))-1))<<bit_partition[i]))>>bit_partition[i])<<
    #     permuted_bit_partition[perm[i]])

    for j in prange(index.size):
        for i in range(len(perm)):
            # bit_range = extract_bit_range(index, bit_partition[i], bit_partition[i+1])
            # insert_bit_range(new_index, bit_range, permuted_bit_partition[perm[i]])
            # Inlined version more efficient:
            new_index[j] |= (
                (
                    index[j]
                    & (
                        ((1 << (bit_partition[i + 1] - bit_partition[i])) - 1)
                        << bit_partition[i]
                    )
                )
                >> bit_partition[i]
            ) << permuted_bit_partition[perm[i]]


def permuter(index, new_index, bit_partition, permuted_bit_partition, perm):
    for i in range(len(perm)):
        # bit_range = extract_bit_range(index, bit_partition[i], bit_partition[i+1])
        # insert_bit_range(new_index, bit_range, permuted_bit_partition[perm[i]])
        # Inlined version more efficient:
        if isinstance(new_index, np.ndarray):
            new_index |= (
                (
                    index
                    & (
                        (int(1 << (bit_partition[i + 1] - bit_partition[i])) - 1)
                        << bit_partition[i]
                    )
                )
                >> bit_partition[i]
            ) << permuted_bit_partition[perm[i]]
        else:
            for j in range(len(index)):
                new_index[j] |= (
                    (
                        index[j]
                        & (
                            (int(1 << (bit_partition[i + 1] - bit_partition[i])) - 1)
                            << int(bit_partition[i])
                        )
                    )
                    >> int(bit_partition[i])
                ) << int(permuted_bit_partition[perm[i]])
            


def invert_perm(perm):
    return [perm.index(i) for i in range(len(perm))]


def permute_axes(index, index_bit_permutation, jit = True):
    n = len(index_bit_permutation)

    index_bit_permutation = -np.array(index_bit_permutation) + n - 1
    index_bit_permutation = list(index_bit_permutation)[::-1]

    index_bit_permutation = list(index_bit_permutation)

    # Get the permutation of the axes

    perm = [index_bit_permutation.index(i) for i in range(n)]

    # Initialize the shape
    perm_shape = n * [1]

    i = 1
    while i < len(perm):
        # If this condition is met, both indices form an interval,
        # that does not need further permutation
        if perm[i] - 1 == perm[i - 1]:
            # Remove the index from the perm and perm_shape list
            perm.pop(i)
            perm_shape.pop(i)

            # Update the shape
            perm_shape[i - 1] += 1
            # Update the permutation indices
            for j in range(len(perm)):
                if perm[j] > perm[i - 1]:
                    perm[j] -= 1
        else:
            i += 1

    if isinstance(index, np.ndarray):
        new_index = np.zeros(index.size, dtype=index.dtype)
        dtype = index.dtype
    else:
        new_index = [0 for _ in range(len(index))]
        dtype = np.int64

    bit_partition = [sum(perm_shape[:i]) for i in range(len(perm_shape) + 1)]

    perm_shape_permuted = [perm_shape[i] for i in invert_perm(perm)]
    permuted_bit_partition = [
        sum(perm_shape_permuted[:i]) for i in range(len(perm_shape) + 1)
    ]

    if len(perm) * len(index) > 2**10 and jit:
        jitted_permuter(
            index,
            new_index,
            np.array(bit_partition),
            np.array(permuted_bit_partition),
            np.array(perm),
        )
    else:
        
        permuter(
            index,
            new_index,
            np.array(bit_partition, dtype = dtype),
            np.array(permuted_bit_partition, dtype = dtype),
            np.array(perm, dtype = dtype),
        )

    return new_index


@njit(cache=True)
def invert_permutation(p):
    """Return an array s with which np.array_equal(arr[p][s], arr) is True.
    The array_like argument p must be some permutation of 0, 1, ..., len(p)-1.
    """
    s = np.empty(len(p), dtype=np.int64)
    s[p] = np.arange(p.size)
    return s


@njit(nogil=True, cache=True)
def bi_array_moveaxis(data_array, index_perm, f_index_array):
    for i in range(len(index_perm)):
        shifted_value = 1 << index_perm[i]
        start_index = 1 << i

        for j in range(start_index):
            f_index_array[start_index ^ j] = shifted_value ^ f_index_array[j]

    # return f_index_array
    return data_array[f_index_array]


@njit(parallel = True, cache = True)
def dense_measurement_brute(input_array, mes_amount, outcome_index):
    
    n = int(np.log2(len(input_array)))
    mes_amount = int(mes_amount)
    reshaped_array = input_array.reshape((2**(mes_amount), 2**(n-mes_amount)))
    
    p_array = np.zeros(2**mes_amount, dtype = input_array.dtype)
    
    for i in prange(2**mes_amount):
        new_array = reshaped_array[i,:]
        p_array[i] = np.vdot(new_array, new_array)
    
    p_array = np.abs(p_array)
    max_p_array = np.max(p_array)
    
    indices = np.nonzero(p_array > max_p_array*cutoff_ratio)[0]
    
    return reshaped_array[indices,:], p_array[indices], indices

    new_arrays = []
    p_values = []
    outcome_indices = []
    
    for i in range(2**mes_amount):
        p = p_array[i]
        if p > max_p_array*cutoff_ratio:
            new_arrays.append(reshaped_array[i, :])
            p_values.append(p)
            outcome_indices.append(outcome_index+i)
    
    return new_arrays, p_values, outcome_indices


@njit(nogil=True, cache=True)
def dense_measurement_smart(input_array, mes_amount, outcome_index):
    
    p = np.abs(np.vdot(input_array, input_array) )
    
    if p < float_tresh:
        return [input_array], [p], [-1]
    
    if mes_amount == 0:
        return [input_array], [p], [outcome_index]
    
    N = input_array.shape[0]
    new_arrays = []
    p_list = []
    outcome_index_list = []
    
    a, b, c = dense_measurement_smart(input_array[:N//2], mes_amount - 1, outcome_index)
    
    if c[0] != -1:
        new_arrays.extend(a)
        p_list.extend(b)
        outcome_index_list.extend(c)
    
    a, b, c = dense_measurement_smart(input_array[N//2:], mes_amount - 1, outcome_index + 2**(mes_amount-1))
    
    if c[0] != -1:
        new_arrays.extend(a)
        p_list.extend(b)
        outcome_index_list.extend(c)
    
    if len(outcome_index_list):
        return new_arrays, p_list, outcome_index_list
    else:
        return new_arrays, p_list, [-1]


@njit(cache = True)
def sort_indices_jitted(row, col, data, shape_1):
    
    shifted_row = row << (int(np.log2(shape_1))+1)
    sorting_array = shifted_row ^ col
    
    array_sort = np.argsort(sorting_array, kind = "mergesort")
    
    new_row = row[array_sort]
    new_col = col[array_sort]
    new_data = data[array_sort]
    
    return new_row, new_col, new_data

@njit(parallel = True, cache = True)
def coo_sparse_matrix_mult_jitted(A_row, A_col, A_data, B_row, B_col, B_data, A_shape, B_shape):
    
    A_row, A_col, A_data = sort_indices_jitted(A_row, A_col, A_data, A_shape[1])
    B_col, B_row, B_data = sort_indices_jitted(B_col, B_row, B_data, B_shape[0])
    
    unique_marker_a = find_unique_markers(A_row)
    unique_marker_b = find_unique_markers(B_col)
    
    res = np.zeros((len(unique_marker_a), len(unique_marker_b)), dtype = A_data.dtype)
    
    for i in prange(len(unique_marker_a)-1):
        comparison_block_a = A_col[unique_marker_a[i]:unique_marker_a[i+1]]        
        for j in range(len(unique_marker_b)-1):
                
            if comparison_block_a[0] > B_row[unique_marker_b[j+1]-1]:
                continue
            
            if B_row[unique_marker_b[j]] > comparison_block_a[-1]:
                continue
            
            comparison_block_b = B_row[unique_marker_b[j]:unique_marker_b[j+1]]
            
            agreeing_ind_a, agreeing_ind_b = find_agreements(comparison_block_a, comparison_block_b)
            
            res[i, j] += np.dot(A_data[unique_marker_a[i] + agreeing_ind_a], B_data[unique_marker_b[j] + agreeing_ind_b])
    
    new_row = []
    new_col = []
    new_data = []
    
    abs_res = np.abs(res)
    max_abs = np.max(abs_res.ravel())
    
    
    for i in range(res.shape[0]):
        for j in range(res.shape[1]):
            if abs_res[i,j] > cutoff_ratio * max_abs:
            # if res[i,j] != 0:
                new_row.append(A_row[unique_marker_a[i]])
                new_col.append(B_col[unique_marker_b[j]])
                new_data.append(res[i,j])
                
    return np.array(new_row, dtype = np.int64), np.array(new_col, dtype = np.int64), np.array(new_data, dtype = A_data.dtype)
                
            
@njit(cache = True)
def find_unique_markers(arr):
    
    unique_marker = [0]
    
    cur_value = arr[0]
    for i in range(len(arr)):
        if cur_value != arr[i]:
            unique_marker.append(i)
            cur_value = arr[i]
            
    unique_marker.append(len(arr))
            
    return unique_marker
    

@njit(cache = True)
def find_agreements(block_a, block_b):

    a_pointer = 0
    b_pointer = 0
    
    a_agreements = []
    b_agreements = []
    
    while a_pointer < len(block_a) and b_pointer < len(block_b):
        
        if block_a[a_pointer] < block_b[b_pointer]:
            a_pointer += 1
        elif block_a[a_pointer] > block_b[b_pointer]:
            b_pointer += 1
        else:
            a_agreements.append(a_pointer)
            b_agreements.append(b_pointer)
            
            a_pointer += 1
            b_pointer += 1
            
    return np.array(a_agreements, dtype = np.int32), np.array(b_agreements, dtype = np.int32)


def coo_sparse_matrix_mult(A, B):
    if A.shape[0] < B.shape[1]:
        new_row, new_col, new_data = coo_sparse_matrix_mult_jitted(A.row, A.col, A.data, B.row, B.col, B.data, A.shape, B.shape)
        
    else:
        
        new_col, new_row, new_data = coo_sparse_matrix_mult_jitted(B.col, B.row, B.data, A.col, A.row, A.data, B.shape[::-1], A.shape[::-1])
        
    if len(new_data) == 0:
        return coo_array(([], ([], [])), shape = (A.shape[0], B.shape[1]))

    if np.max(new_row) < 1<<31 and not new_row.dtype == np.int32:
        new_row = np.array(new_row, dtype = np.int32)
        
    if np.max(new_col) < 1<<31 and not new_col.dtype == np.int32:
        new_col = np.array(new_col, dtype = np.int32)
            
    return coo_array((new_data, (new_row, new_col)), shape = (A.shape[0], B.shape[1]))


def sparse_matrix_mult(A, B):
    
    if A.shape[0]*A.shape[1] < B.shape[0]*B.shape[1]:

        log_shape_0_a = np.log2(A.shape[0])
        log_shape_1_b = np.log2(B.shape[1])
        
        log_sparsity_a = -np.log2(A.nnz/(A.shape[0]*A.shape[1]))
        log_sparsity_b = -np.log2(B.nnz/(B.shape[0]*B.shape[1]))
        
    else:
        log_shape_0_a = np.log2(B.shape[1])
        log_shape_1_b = np.log2(A.shape[0])
        
        log_sparsity_a = -np.log2(B.nnz/(B.shape[0]*B.shape[1]))
        log_sparsity_b = -np.log2(A.nnz/(A.shape[0]*A.shape[1]))
    
    if get_prediction(log_shape_0_a, log_shape_1_b, log_sparsity_a, log_sparsity_b) and False:
        return (A @ B).tocoo()
    else:
        return  coo_sparse_matrix_mult(A, B)
    

def get_prediction(log_shape_0_a, log_shape_1_b, log_sparsity_a, log_sparsity_b):
    
    data_tuple = (log_shape_0_a, log_shape_1_b, log_sparsity_a, log_sparsity_b)
    
    # Normalize and add bias to the input data
    x_mean = np.array([ 5.488, 28.741,  7.679, 26.106])
    x_std = np.array([ 2.26844793, 11.88553402,  3.73764083, 12.94182228])
    theta = np.array([-3.37310532,  1.05192584, -0.98338696, -0.6838338 , -2.10351854])
    normalized_data = (np.array(data_tuple) - x_mean) / x_std
    data_with_bias = np.hstack(([1], normalized_data))

    # Sigmoid function
    def sigmoid(z):
        return 1 / (1 + np.exp(-z))
    
    # Calculate the probability using the trained theta values
    probability = sigmoid(np.dot(data_with_bias, theta))

    # Make a binary prediction (True/False) based on the probability threshold (0.5)
    prediction = probability >= 0.5

    return prediction    


# imports 
from qrisp.qiro.qiro_problem import QIROProblem
from qrisp.qaoa.problems.maxSatInfrastr import maxSatclCostfct, clausesdecoder
from qrisp.qiro.qiroproblems.qiroMaxSatInfrastr import * 
from qrisp.qiro.qiro_mixers import qiro_init_function, qiro_RXMixer
from qrisp import QuantumVariable
import matplotlib.pyplot as plt
import networkx as nx

def test_QIROmaxSat():
    # define the problem according to maxSat encoding
    problem = [8 , [[1,2,-3],[1,4,-6], [4,5,6],[1,3,-4],[2,4,5],[1,3,5],[-2,-3,6], [1,7,8], [3,-7,-8], [3,4,8],[4,5,8], [1,2,7]]]
    decodedClauses = clausesdecoder( problem)

    qarg = QuantumVariable(problem[0])

    # set simulator shots
    mes_kwargs = {
        #below should be 5k
        "shots" : 5000
        }


    # assign cost_function and maxclique_instance, normal QAOA
    testCostFun = maxSatclCostfct(problem)

    # assign the correct new update functions for qiro from above imports
    qiro_instance = QIROProblem(problem = problem,  
                                replacement_routine = create_maxSat_replacement_routine, 
                                cost_operator = create_maxSat_cost_operator_reduced,
                                mixer = qiro_RXMixer,
                                cl_cost_function = maxSatclCostfct,
                                init_function = qiro_init_function
                                )


    # We run the qiro instance and get the results!
    res_qiro = qiro_instance.run_qiro(qarg=qarg, depth = 3, n_recursions = 2, 
                                    #mes_kwargs = mes_kwargs
                                    )



    import itertools
    def testCostFun(state):
        obj = 0
        #literally just do a minus 1 op if state is equiv to a given condition
        for index in range(len(decodedClauses)):
            if state in decodedClauses[index]:
                obj -= 1 

        return obj


    #print 5 best sols 
    # assert that sol is in decoded clauses 
    maxfive = sorted(res_qiro, key=res_qiro.get, reverse=True)[:5]
    # assert that sol is in decoded clauses 
    for name in res_qiro.keys():  # for name, age in dictionary.iteritems():  (for Python 2.x)
        if name in maxfive:
            print(name)
            print(testCostFun(name))
        if testCostFun(name) < 0:
            temp = False
            for index in range(len(decodedClauses)):
                if name in decodedClauses[index]:
                    temp  = True
            assert temp




"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp.qtypes import QuantumFloat
from qrisp.arithmetic.comparisons import less_than
from qrisp.arithmetic.modular_arithmetic.mod_tools import modinv, montgomery_encoder
from qrisp.environments import custom_control
from qrisp import cx, bin_rep, control, swap, invert, redirect_qfunction, QuantumModulus, mcx, fast_append, auto_uncompute
# This file implements the techniques described in this paper: https://arxiv.org/abs/1801.01081
# The goal is to have performant modular multiplication. To this end, instead of taking the
# much explored path of creating a modular adder, that is used within the multiplication,
# the paper instead describes multiple techniques to perform the modulo reduction
# after the (non-modular) multiplication has concluded.

# In this file we implement the described techniques for Montgomery reduction.

def cl_montgomery_reduction(t, N, m):
    """
    This function serves as an example, how classical Montgomery reduction works.
    
    Note, that the return value is not t%N but (t*2**-m)%N. This is called Montgomery
    form and doesn't really create in any problems because it is just a matter of
    encoding.

    Parameters
    ----------
    t : integer
        The value to be reduced.
    N : integer
        The modulus.
    m : integer
        The Montgomery shift. Needs to satisfy t < N 2**n.

    Returns
    -------
    S : integer
        The result of the reduction: S = (t*2**-m)%N
    u : integer
        A garbage variable with the value u = (-tN**-1)%(2**m)

    """
    
    S = t
    u = int(0)
    for k in range(m):
        temp = S%2
        S -= temp*N
        S = S//2
        
        u = u | temp<<k
    
    if S < 0:
        S += N
    
    return S, u



# This function transfer the least significant bit from one QuantumFloat to another one
def transfer_lsb(from_qv, to_qv):
    
    # Get the qubit in question
    lsb = from_qv.reg.pop(0)
    
    # Adjust the relevant QuantumFloat attributes
    from_qv.size -= 1
    from_qv.msize -= 1
    from_qv.mshape[0] += 1
    from_qv.exponent += 1

    # Insert the qubit in the target
    to_qv.reg.insert(len(to_qv), lsb)
    
    # Adjust the relevant attributes
    to_qv.size += 1
    to_qv.msize += 1
    to_qv.mshape[1] += 1


# This function realizes the first step of the Montgomery reduction

def QREDC(t, N, m):
    # Create the variable holding u
    u = QuantumFloat(0, qs = t.qs, name = "u*")
    
    # Set an alias for S
    S = t
    
    # Perform the loop of the Montgomery reduction
    for k in range(m):
        
        # Set the alias similar to the classical version
        temp = S[0]
        
        # Transfer the least significant bit to u
        transfer_lsb(S, u)
        
        
        # Perform the step S -= temp*N
        
        # The subtraction only happens if S is odd. Since N
        # is odd by design, the new least significant qubit would always
        # ends up to be zero (this is why we can do the S = S/2 step)
        # Since we already transfered the least significant qubit to u,
        # we would have to create a new qubit to perform the subtraction
        # where we however know that this qubit will end up to be zero and
        # it would also immediately phased out again by the S = S/2 step.
        # Therefore instead of subtracting N, we subtract (N-1) on 
        # "truncated" S. 
        
        # To accomodate for the fact, that S is "truncated" ie. currently it's
        # least significant qubit has significance 2**1 = 2, we additionally divide
        # (N-1) by two.
        with control(temp):
            t.inpl_adder((-(N-1)//2), S)
        
        # Perform the S = S/2 step.        
        S.exp_shift(-1)
    
    
    
    # Perform the correction step
    # if S < 0:
    #     S += N
    with control(S.sign()):
        t.inpl_adder(N, S[:-1])
        
    
    # The following steps are not part of the original Montgomery reduction algorithm
    # but as described in the paper, they help with the uncomputation of u
    cx(S[0], S.sign())
    
    # Transfer the sign to u
    sgn = S.reg.pop(-1)
    S.size -= 1
    S.signed = False
    
    u.reg.insert(len(u), sgn)
    u.size += 1
    u.mshape[1] += 1
    
    # Adjust the m attribute, which indicates the current Montgomery shift of this
    # QuantumModulus
    S.m = S.m - m
    
    return S, u


# This function performs Montgomery reduction and subsequently uncomputes the u value
def montgomery_red(t, a, b, N, m, permeable_if_zero = False):
    
    # Perform Montgomery reduction
    t, u = QREDC(t, N, m)
    
    # Perform the uncomputation as described in the paper
    for k in range(len(a)):
        with control(a[k]):
            t.inpl_adder(-((2**k*b))*modinv(N, 2**(m+1)), u)
    
    if permeable_if_zero:    
        # cx(t[0], u[-1])
        pass
        # mcx([a[-1]] + [t[0]], u[-1], ctrl_state = "01")
        mcx(list(a) + [t[0]],  
            u[-1], ctrl_state = "0"*len(a) + "1",
            method = "balauca")
    
    # Delete the uncomputed value
    u.delete(verify = False)
    return t


# This function searches for a suitable Montgomery shift m
# On one hand, we want to have m as small as possible (because it requires extra
# qubits and iterations in the Montgomery reduction) but on the other hand we
# are limited by the requirement that a*b = t < N*2**m
def find_best_montgomery_shift(b, N):
    # The idea is now to try out increasingly large m, compute the corresponding
    # Montgomery form of the (classical) multiplication value b and check, if it
    # satisfies the above condition
    
    m = 0
    while True:
        b_trial = montgomery_encoder(int(b), 1<<m, N)%N
        # Since a is reduced, in the worst case, it is equal to N-1
        if (N-1)*b_trial < N*2**m:
            return m
        m += 1


# This function performs semi classical out of place multiplication follow by a
# Montgomery reduction
def montgomery_mod_semi_mul(a, b, output_qg = None, permeable_if_zero = False):
    
    # Set some aliases
    N = a.modulus
    
    # Reduce b
    b = b%N
    
    # Check some special cases that can be treated efficiently
    if b == 0:
        if output_qg is None:
            return a.duplicate()
        else:
            return output_qg
    if b == 1:
        if output_qg is None:
            return a.duplicate(init = True)
        else:
            a.inpl_adder(a, output_qg)
            return output_qg
    
    # Determine a suitable Montgomery shift
    m_shift = find_best_montgomery_shift(b, N)
    
    # Encode the Montgomery shift
    b_encoded = montgomery_encoder(b, 2**m_shift, N)%N
    
    # The idea here is to multiply the encoded b instead of original b
    # After Montgomery reduction, we have
    # (a*b_encoded*2**-m_shift)%N
    # = (a*b*2**m_shift*2**-m_shift)%N
    # = (a*b)%N
    
    # If no output value is given, create it
    if output_qg is None:
        t = QuantumFloat(a.size+m_shift, signed = True, qs = a.qs)
    else:
        if output_qg.modulus != N:
            raise Exception("Output QuantumModulus has incompatible modulus")
        
        # Extend the output to hold the overflow of the multiplication
        output_qg.extend(m_shift, 0)
        
        # Add the sign
        output_qg.add_sign()
        
        # Set the alias
        t = output_qg
        
    
    
    # Perform the non-modular multiplication
    
    n = int(np.ceil(np.log2(N)))
    b_encoded_string = bin_rep(b_encoded, n)[::-1]
    for i in range(n):
        if b_encoded_string[i] == "1":
            a.inpl_adder(a, t[i:])
    
    # Typecast to QuantumModulus and increase the Montgomery shift    
    from qrisp import QuantumModulus
    t.__class__ = QuantumModulus
    t.modulus = a.modulus
    t.m = a.m + m_shift
    t.inpl_adder = a.inpl_adder
    
    # Perform Montgomery reduction and return
    return montgomery_red(t, a, b_encoded, N, m_shift, permeable_if_zero = permeable_if_zero)


@custom_control
def ft_swap(a, b, ctrl = None):
    
    if ctrl is None:
        swap(a,b)
    else:
        cx(a, b)
        mcx([ctrl, b], a, method = "jones")
        cx(a, b)


# This function perform semi-classical in-place multiplication
@custom_control
def semi_cl_inpl_mult(a, X, ctrl = None, treat_invalid = False):
    
    # The idea here is to perform two semi-classical out of place multiplications,
    # where the second one uncomputes the input value.
    # After that, the variables are swapped, such that the input variable holds the
    # result and the temporary variable holds |0> and thus can be deleted.
    
    # An additional feature of this function is that it can be efficiently controlled.
    # We achieve this by controlling the swaps. If the control argument contains a 
    # QuantumBool that is set to True, another swap will be executed between the temporary
    # register and the input register.
    # This results in the fact that the first out of place multiplication will have 
    # no effect, because the input value of this multiplication is 0.
    # Before the second out of place multiplication happens, we perform another swap,
    # such that the input value of this multiplication is 0 as well.
    # Finally, we swap again, to restore the original input value into the input register.
    
    # Perform a reduction
    X = X%a.modulus
    
    
    # Check some special cases
    if X == 0:
        raise Exception("Tried to perform in-place multiplication with 0 (not invertible)")
    if X == 1:
        return a

    
    with fast_append(2):
        
        # Create the temporary value    
        tmp = a.duplicate(qs = a.qs)
        
        # If treat_invalid is set to True, the function should leave the invalid values
        # invariant. That is, values that are bigger than the modulus N.
        
        # We achieve this by computing a QuantumBool, which indicates wether the value is
        # invalid and the controlling the multiplication on this QuantumBool.
        if treat_invalid:
            a.__class__ = QuantumFloat
            reduced = a < a.modulus
            a.__class__ = QuantumModulus
            ctrl = [ctrl, reduced[0]]
        
        # If the controlled version of this function is required, we perform the swapping
        # strategy from above.
                
        if ctrl is not None:
            with control(ctrl, invert = True):
                for i in range(len(a)):
                    # swap(tmp[i], a[i])
                    ft_swap(tmp[i], a[i])


        
        # Perform the out of place multiplication
        tmp = montgomery_mod_semi_mul(a, 
                                      X, 
                                      output_qg = tmp,
                                      permeable_if_zero = ctrl is not None)
        
        # Perform the intermediate swap
        if ctrl is not None:
            with control(ctrl, invert = True):
                for i in range(len(a)):
                    # swap(tmp[i], a[i])
                    ft_swap(tmp[i], a[i])
        
        # The second out of place multiplication is the quantum inverted version
        # of the multiplication with the modular inverse
        # Why does this uncompute the input value?
        # Before the first out of place multiplication, we are in the state
        # |a>|0>
        # After the multiplication, we have
        # |a>|(a*X)%N>
        # If we rewrite this with b = (a*X)%N, we have
        # |(b*X**-1)%N>|b>
        # This could also be seen as the result of an out of place multiplication
        # from the tmp variable into the input variable by the number X**-1.
        # We can use this insight to uncompute the input variable by simply reverting
        # precisely this operation.
        
        with invert():
            a = montgomery_mod_semi_mul(tmp, 
                                        modinv(X, a.modulus), 
                                        output_qg = a, 
                                        permeable_if_zero = ctrl is not None)
        
        # Perform the corresponding swaps
        if ctrl is not None:
            with control(ctrl, invert = False):
                for i in range(len(a)):
                    # swap(tmp[i], a[i])
                    ft_swap(tmp[i], a[i])
        else:
            for i in range(len(a)):
                swap(tmp[i], a[i])
        
        # Delete the temporary variable
        tmp.delete(verify = False)
        
        # Uncompute the reduced bool    
        if treat_invalid:
            a.__class__ = QuantumFloat
            redirect_qfunction(less_than)(a,a.modulus,target= reduced)
            a.__class__ = QuantumModulus
            reduced.delete(verify = True)
        
        return a


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


# Modification to Kahns algorithm to reduce the depth of general circuits by applying
# trivial and non-trivial commutation relations based on the dag representation of
# unqomp.

import networkx as nx
import numpy as np
from numba import njit

from qrisp.uncomputation.unqomp import dag_from_qc

# The following code aims to represent quantum circuits as an array of integers.
# The idea is here that in a 6 qubit quantum circuit, a gate that is executed 
# on the qubits 1 and 3 is represented by the number 001010 = 5 (ie. the digits
# with significance 2**1 and 2**3 are set to 1).
# This construction might remove some of the information (ie. gate type or qubit order)
# but allows efficient processing since, the quantum circuit can be given to numba
# as a numpy array.

# Unfortunately, numpy integers can only represent 64 bit numbers. This means,
# that this technique could only handle quantum circuits with up to 64 qubits.
# We therefore SPLIT the number in the following way:
# The qubits from 0 to 63 represented by a numpy int,
# the qubits from 64 to 127 are represented by another numpy int,
# the qubits from 128 to 255 are represented by another numpy int,
# etc.
# We therefore represent the quantum circuit with a two-index numpy array
# where the first index indicates which gate is described and the second index
# indicates what qubit range is meant.


# This function receives the indices of the qubits of a gate and turns them into
# the list according to the above construction. The parameter k indicates 
# how many integers the result should contain.
def split_integer(digit_list, k):
    
    digit_list.sort()
    
    if k <= 0:
        raise ValueError("Parameter k must be greater than zero")

    result = [0]*k
    i = 0
    for d in digit_list:
        while d//64 > i:
            i += 1
        result[i] |= (1<<(d%64))
        if result[i] >= 1<<64:
            raise
            
    return result

# This function receives a list of qubits and a quantum circuit and turns it into
# a list according to the above construction
def qb_set_to_int(qubits, index_dict):
    qb_indices = [index_dict[qb] for qb in qubits]
    return split_integer(qb_indices, int(np.ceil(len(index_dict)/64)))

# This function reverses the split_integer function (usefull for debugging)
def reverse_split(split_int, num_qubits):
    
    res = []
    for i in range(num_qubits):
        if not i%64:
            current_partial_int = split_int.pop(0)
        
        if (current_partial_int) & (1<<(i%64)):
            res.append(i)
            
    return res

# This function performs the parallelization. As described above, the idea is to
# build up the DAG from the quantum circuit and determine a linearization, which
# is in-turn used to achieve a more optimal quantum circuit.
# The linearization is performed using a modified version of Kahn's algorithm.
# This modification is informed about the depth of the circuit because of the above 
# constructions.
def parallelize_qc(qc, depth_indicator = None):
    if len(qc.data) <= 1:
        return qc.copy()
    
    if depth_indicator is None:
        depth_indicator = lambda x : 1

    dag = dag_from_qc(qc, remove_init_nodes=True)

    sprs_mat = nx.to_scipy_sparse_array(dag, format="csr")

    node_list = list(dag.nodes())
    
    # This list will contain the participating qubits of each gate in the above
    # discussed representation
    qubit_ints = []
    
    # This list will contain the depth of each gate, which can be specified via
    # the depth indicator function.
    depth_indicators = []
    
    index_dict = {qc.qubits[i] : i for i in range(len(qc.qubits))}
    
    for n in node_list:
        qubit_ints.append(qb_set_to_int(n.instr.qubits, index_dict))
        depth_indicators.append(depth_indicator(n.instr.op))
    
    # Convert to array
    qubit_ints = np.array(qubit_ints, dtype = np.uint64)

    # Execute topological sort
    res = depth_sensitive_topological_sort_jitted(
        sprs_mat.indices, sprs_mat.indptr, qubit_ints, num_qubits=qc.num_qubits(), depth_indicators = np.array(depth_indicators)
    )

    # Build new circuit
    qc_new = qc.clearcopy()

    for i in range(len(res)):
        qc_new.append(node_list[res[i]].instr)

    return qc_new



# Kahns Algorithm based on
# https://www.geeksforgeeks.org/topological-sorting-indegree-based-solution/
def depth_sensitive_topological_sort(indices, indptr, int_qc, num_qubits, depth_indicators):
    # Create a vector to store indegrees of all
    # vertices. Initialize all indegrees as 0.
    n = len(indptr) - 1
    in_degree = np.zeros(n, dtype=np.int32)

    depths = np.zeros(num_qubits, dtype=np.int32)
    # Traverse adjacency lists to fill indegrees of
    # vertices.  This step takes O(V + E) time

    for i in range(n):
        for j in indices[indptr[i] : indptr[i + 1]]:
            in_degree[j] += 1

    # Create a queue and enqueue all vertices with
    # indegree 0
    queue = []

    for i in range(n):
        if in_degree[i] == 0:
            queue.append(i)

    # Initialize count of visited vertices
    cnt = 0

    # Create a vector to store result (A topological
    # ordering of the vertices)
    top_order = np.zeros(n, dtype=np.int32)

    # One by one dequeue vertices from queue and enqueue
    # adjacents if indegree of adjacent becomes 0

    while queue:
        # The depth sensitive part is now to deque the node with the least depth
        node_costs = np.zeros(len(queue))

        for i in range(len(queue)):
            node = queue[i]

            qubits = int_qc[node]

            # Collect the depth of each participating qubit
            depth_list = []
            for j in range(num_qubits):
                if qubits[j//64] & (1 << (j%64)):
                    depth_list.append(depths[j])
            
            # If multiple gates have the same max depth, the faster ones should
            # be executed first, because they might block other gates
            node_costs[i] = np.max(np.array(depth_list)) + depth_indicators[node]/10**8

        u = queue.pop(np.argmin(node_costs))

        top_order[cnt] = u

        # Update depths array
        max_depth = 0

        for i in range(num_qubits):
            if int_qc[u, i//64] & 1 << (i%64):
                if depths[i] > max_depth:
                    max_depth = depths[i]

        for i in range(num_qubits):
            if int_qc[u, i//64] & 1 << (i%64):
                depths[i] = max_depth + depth_indicators[u]

        # Update in degree array
        for i in indices[indptr[u] : indptr[u + 1]]:
            in_degree[i] -= 1
            if in_degree[i] == 0:
                queue.append(i)

        cnt += 1

    return top_order


depth_sensitive_topological_sort_jitted = njit(cache=True)(
    depth_sensitive_topological_sort
)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

def find_max(single_cor, double_cor, res, solutions):
    """
    General subroutine for finding the values with maximal correlation in the QIRO algorithm

    Parameters
    ----------
    single_cor : List
        Single qubits correlations to check.
    double_cor : List[Tuple]
        Multi qubits correlations to check.
    res : dict
        Result dictionary of initial QAOA optimization procedure
    solutions : List
        Qubits which have been found to be positive correlated, i.e. part of the problem solution  

    Returns
    -------
    max_item , sign
        The item with maximal correlation and the sign of the correlation.

    """
    
    max = 0

    for item2 in double_cor:
        if abs(item2[0]) == abs(item2[1]):
            continue
        summe = 0

        # calc correlation expectation
        for key, val in res.items():
            summe += pow(val, 2) * pow(-1, int(key[int(abs(item2[0]))])) * pow(-1, int(key[int(abs(item2[1]))]))

        #find max
        if abs(summe) > abs(max):
            max, max_item = summe, item2
            sign = np.sign(summe)

    for node in single_cor:
        if node in solutions:
            continue

        for key, val in res.items():
            summe += val * pow(-1, int(key[int(node)])) 
            
        if abs(summe) > abs(max):
            max, max_item = summe, node
            sign = np.sign(summe)
    
    return max_item, sign


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import itertools
from qrisp.qaoa import QAOAProblem
from qrisp.qaoa.mixers import RZ_mixer
from qrisp.qaoa.problems.minSetCoverInfrastr import minSetCoverclCostfct,minSetCoverCostOp, init_state

from qrisp import QuantumVariable

def test_QAOAminSetCover():


    """
    This is an example of the minSetCoer QAOA solution
    the problem is structured as follows:
    > Given a universe [n] and m subsets S (S_j in S, S_j is subset of [n]), find the minimum cardinality subcollection S' (S' is subset of S) s.t. their union recovers [n]. 

    We will not stick to mathematical assignment of variable names. 

    """


    # sets are given as list of lists
    examplesets = [[0,1,2,3],[1,5,6,4],[0,2,6,3,4,5],[3,4,0,1],[1,2,3,0],[1]]
    # full universe is given as a tuple
    sol = (0,1,2,3,4,5,6)
 

    # assign operators
    cost_fun = minSetCoverclCostfct(sets=examplesets,universe = sol)
    mixerOp = RZ_mixer
    costOp = minSetCoverCostOp(sets=examplesets, universe=sol)

    #initialize variable
    qarg = QuantumVariable(len(examplesets))

    #+run qaoa
    QAOAinstance = QAOAProblem(cost_operator=costOp ,mixer= mixerOp, cl_cost_function=cost_fun)
    QAOAinstance.set_init_function(init_function=init_state)
    InitTest = QAOAinstance.run(qarg=qarg, depth=5)

    # create example cost_func
    def testCostFun(state,universe):
        obj = 0
        #literally just do a minus 1 op if state is equiv to a given condition
        intlist = [s for s in range(len(list(state))) if list(state)[s] == "1"]
        sol_sets = [examplesets[index] for index in intlist]
        res = ()
        for seto in sol_sets:
            res = tuple(set(res+ tuple(seto))) 
        if res == universe:
            obj += len(intlist)

        return obj

    # test if cost_func functions for all items . 
    for state in InitTest.keys():
        if testCostFun(state=state, universe = sol) > 0:
            intlist = [s for s in range(len(list(state))) if list(state)[s] == "1"]
            sol_sets = [examplesets[index] for index in intlist]
            res = ()
            for seto in sol_sets:
                res = tuple(set(res+ tuple(seto))) 
            assert sol == res

    

    #print the ideal solutions
    #print("5 most likely Solutions") 
    maxfive = sorted(InitTest, key=InitTest.get, reverse=True)[:5]
    for name in InitTest.keys():  # for name, age in dictionary.iteritems():  (for Python 2.x)
        if name in maxfive:
            print(name)
            print(testCostFun(name, universe=sol))  

    #find ideal solution by brute force    
    temp_binStrings = list(itertools.product([1,0], repeat=len(examplesets)))
    binStrings = ["".join(map(str, item))  for item in temp_binStrings]
    
    mini = len(examplesets)
    for index in range(len(binStrings)):
        val = testCostFun(binStrings[index], universe=sol)
        if 0 < val < mini :
            
            mini = val
            min_index = index

    optimal_sol = binStrings[min_index]
    print("optimal solution")
    print(optimal_sol)


    # write approximation ratio test
    # metric_dict ={}
    # metric_dict["counts"] = InitTest
    # metric_dict["cost_func"] = minSetCoverclCostfct(sets=examplesets,universe = sol)
    # metric_dict["optimal_sol"] = optimal_sol
    
    # print(approximation_ratio(metric_dict))
    #assert 0 < approximation_ratio(metric_dict)<= 3  

    import numpy as np

    randInteg = np.random.randint(4,8)
    
    randInteg2 = np.random.randint(0,1)
    
    sol = tuple(range(randInteg))
    if randInteg2 == 0:
        examplesets = [sol,sol[0:randInteg-3],sol[randInteg-3:randInteg-1],sol[randInteg-1:]] 
    else:
        print(randInteg)
        examplesets = [[index] for index in range(randInteg)]

    # assign the operators
    cost_fun = minSetCoverclCostfct(sets=examplesets,universe=sol)
    mixerOp = RZ_mixer
    costOp = minSetCoverCostOp(sets=examplesets, universe=sol)

    #initialize the qarg
    qarg2 = QuantumVariable(len(examplesets))

    # run the qaoa
    QAOAinstance = QAOAProblem(cost_operator=costOp ,mixer= mixerOp, cl_cost_function=cost_fun)
    QAOAinstance.set_init_function(init_function=init_state)
    InitTest = QAOAinstance.run(qarg=qarg2, depth=5)

    maxOne =  sorted(InitTest, key=InitTest.get, reverse=True)[:1]
    if randInteg2 == 0:
        print(maxOne)
        assert maxOne == ["1000"]
    else:
        print(maxOne)
        assert maxOne == ["1"* len(examplesets)]




# imports 
from qrisp.qiro.qiro_problem import QIROProblem
from qrisp.qaoa.qaoa_problem import QAOAProblem
from qrisp.qaoa.problems.maxCliqueInfrastr import maxCliqueCostfct, maxCliqueCostOp
from qrisp.qiro.qiroproblems.qiroMaxCliqueInfrastr import * 
from qrisp.qaoa.mixers import RX_mixer
from qrisp.qiro.qiro_mixers import qiro_init_function, qiro_RXMixer
from qrisp import QuantumVariable
import matplotlib.pyplot as plt
import networkx as nx

def test_QIRO():

    # First we define a graph via the number of nodes and the QuantumVariable arguments
    num_nodes = 15
    G = nx.erdos_renyi_graph(num_nodes,0.7, seed =  99)
    Gtwo = nx.erdos_renyi_graph(num_nodes,0.7, seed =  99)
    qarg = QuantumVariable(G.number_of_nodes())
    # set simulator shots
    mes_kwargs = {
        #below should be 5k
        "shots" : 5000
        }

    #assign cost_function and maxclique_instance, normal QAOA
    testCostFun = maxCliqueCostfct(Gtwo)
    maxclique_instance = QAOAProblem(maxCliqueCostOp(G), RX_mixer, maxCliqueCostfct(G))

    # assign the correct new update functions for qiro from above imports
    qiro_instance = QIROProblem(problem = Gtwo,  
                                replacement_routine = create_maxClique_replacement_routine, 
                                cost_operator = create_maxClique_cost_operator_reduced,
                                mixer = qiro_RXMixer,
                                cl_cost_function = maxCliqueCostfct,
                                init_function = qiro_init_function
                                )


    res_qiro = qiro_instance.run_qiro(qarg=qarg, depth = 3, n_recursions = 2, 
                                    #mes_kwargs = mes_kwargs
                                    )

    #QIRO 5 best results
    maxfive = sorted(res_qiro, key=res_qiro.get, reverse=True)[:5]

    import itertools
    for key, val in res_qiro.items():  
        if key in maxfive:
            cost = testCostFun({key:1})
            temp = True
            if cost < 0:
                intlist = [s for s in range(len(list(key))) if list(key)[s] == "1"]
                combinations = list(itertools.combinations(intlist, 2))
                # if any combination is found in the list of G.edges(), the solution is wrong, and energy == 0
                for combination in combinations:
                    if combination not in G.edges():
                        temp = False
                        break
            assert temp




"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import QuantumVariable,  mcx, rz, x, rx 

from qrisp import control
from collections.abc import Iterable
# this is pretty much maxIndependent set, but with a twist
# instead of swapping out singular qubits, you swap out whole predefined sets. 
# this means you apply the mixers to all elements in the sets

# we have a graph of 9 vertices

#######################
## reformulate using @auto_uncompute !!!
## https://www.qrisp.eu/reference/Core/Uncomputation.html




def maxSetPackCostOp(problem):
    
    """
    |  Create the cost/problem operator for this problem instance. The swapping rule is to swap a set in and out of the solution, if it is not intersecting with any other set.
    |  Idea - Per set: 

    * Create ancillas for every element, they represent these elements
    * Perform multi controlled x operations on each ancilla
    * Controls are given by sets with also contain the considered element
    * If all controls are "0" (see ``ctrl_state`` for ``mcx``-operation) we set this ancilla to "1"

    |  Then perform mcx on the qubit describing the set as follows:
    |  If all ancillas are "1" this means the qubits describing the sets contain no intersections with the considered set. We can then swap the set in (or out).
    |  Afterwards uncompute the ancillas.

    Parameters
    ----------
    sets : list(Lists)
        The sets the universe is seperated into as by the problem definition

    universe: Tuple
        The universe for the problem instance, i.e. all possible values (all graph vertices)

    Returns
    -------
    QuantumCircuit: qrisp.QuantumCircuit
        the Operator applied to the circuit-QuantumVariable

    Examples
    --------
    Definition of the sets, given as list of lists. Full universe ``sol`` is given by the amount of elements (+1, since its 0-indexed)
    
    >>> sets = [[0,7,1],[6,5],[2,3],[5,4],[8,7,0],[1]]
    >>> sol = 9
    >>> problem = [sol, sets]

    The relations between the sets, i.e. which vertice is in which other sets

    >>> print(get_neighbourhood_relations(sets, len_universe=len(sol)))

    Assign the operators

    >>> cost_fun = maxSetPackclCostfct(problem)
    >>> mixerOp = RZ_mixer
    >>> costOp = maxSetPackCostOp(problem)
    """

    universe = list(range(problem[0]))
    sets = problem[1]
    
    if not isinstance(sets, Iterable):
        raise Exception("Wrong structure of problem - clauses have to be iterable!")
    for clause in sets:
        if not isinstance(clause, Iterable):
            raise Exception("Wrong structure of problem - each set has to be a tuple!")
        for item in clause:
            if not isinstance(item, int):
                raise Exception("Wrong structure of problem - each literal has to an int!")

    # get neigbhourhood relations from helper function
    nbh_rel = get_neighbourhood_relations(problem)


    def theCostOpEmbedded(qv, gamma):
        #check all sets
        for set_index in range(len(sets)):
            # get set elements and create an ancilla for every set element
            nodes = sets[set_index]
            ancillas = QuantumVariable(len(nodes))
            # go through all ancillas and, equivalently set elements
            for ancilla_index in range(len(ancillas)):
                # if the element is only in one set, we can set this ancilla to 1
                if len(nbh_rel[nodes[ancilla_index]])<2:
                    x(ancillas[ancilla_index])
                    continue
                # else save the sets with also contain the considered element
                nbh_sets_list = [ item for item in nbh_rel[nodes[ancilla_index]] if item != set_index]
                # perform mcx on ancilla, control given by the relevant set
                mcx([qv[nbh_sets_index] for nbh_sets_index in nbh_sets_list], ancillas[ancilla_index], ctrl_state= "0" * len(nbh_sets_list))
            # perform mcrx gate on the qubit describing the considered set
            with control(ancillas):
                rx(gamma, qv[set_index])  

            ancillas.uncompute()

    return theCostOpEmbedded





def get_neighbourhood_relations(problem):
    """
    helper function to return a dictionary describing neighbourhood relations in the sets, i.e. for each element in the universe, gives the info in which the element is contained in.


    Parameters
    ----------
    problem : List 
        The problem definition, as described above

    Returns
    -------
    neighbourhood relation dictionary :  dict
        |  keys: all universe elements (all graph vertices)
        |  values: per universe element the sets it is contained in

    """

    sets = problem[1]
    
    n_dict = {}
    for index_node in range(problem[0]):
        adding_list = [index_set for index_set in range(len(sets)) if index_node in sets[index_set]]
        #if len(adding_list)>1: 
        n_dict[index_node] = adding_list
    return n_dict



def maxSetPackclCostfct(problem):

    """
    create the classical cost function for the problem instance

    Parameters
    ----------
    problem : List 
        The problem definition, as described above

    Returns
    -------
    Costfunction : function
        the classical function for the problem instance, which takes a dictionary of measurement results as input
    """

    universe = list(range(problem[0]))
    sets = problem[1]

    def setupaClCostfct(res_dic):
        energy = 0
        total_counts = 0
        for state in list(res_dic.keys()):
            # assume solution is correct
            list_universe = [True]*len(universe)
            temp = True
            obj = 0
            #get all sets marked by the solution
            intlist = [s for s in range(len(list(state))) if list(state)[s] == "1"]
            sol_sets = [sets[index] for index in intlist]
            
            for seto in sol_sets:
                for val in seto:
                    if list_universe[val]:
                        # if the value appears in the sets set this value to false
                        list_universe[val] = False
                    else: 
                        # is the value is False this element appeared in another solution set
                        # the sets then intersect and the solution is wrong
                        temp = False 
                        break
            if temp:
                obj -= len(intlist)
            energy += obj * res_dic[state]
            total_counts += res_dic[state]
        #print(energy/total_counts)

        return energy/total_counts
    
    return setupaClCostfct


def init_state(qv):
    # all in 0
    return qv





"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.qaoa import QAOAProblem
from qrisp.qaoa.problems.eThrTwoLinInfrastr import eTwoThrLinclCostfct, eThrTwocostOp, init_state
from qrisp.qaoa.mixers import RX_mixer 
from qrisp import QuantumVariable
import numpy as np 


def test_eTwoThrLinQAOA():
    # first three are the indices (a_1,j ; a_2,j ; a3,j), last one is b_L
    #clauses = [(1,3,4,5), (2,6,9,11), (1,5,8,33), (3,4,5,77), (2,4,6,7), (1,3,4,55),(1,3,6,49),(2,4,7,87),  (1,3,4,55),(1,3,6,43),(2,4,7,89), (1,3,4,57),(1,3,6,44),(2,4,7,8)]
    for index in range(10):
        clauses = []
        for index2 in range(4):
            index2a = np.random.randint(0,1)
            if index2a % 2 == 0:
                int1 = np.random.randint(3,5)
                int2 = np.random.randint(0,3)
                int3 = np.random.randint(5,7)
                int4 = np.random.randint(30,39)
                clauses.append((int1,int2,int3,int4))
            else:
                int1 = np.random.randint(1,4)
                int2 = np.random.randint(0,1)
                int3 = np.random.randint(4,7)
                int4 = np.random.randint(31,40)
                clauses.append((int1,int2,int3,int4))

        #assign cost function
        cl_cost_function=eTwoThrLinclCostfct(clauses)
        print(clauses)

        #assign qarg
        qarg = QuantumVariable(7)

        # run the qaoa
        QAOAinstance = QAOAProblem(cost_operator=eThrTwocostOp(clauses=clauses), mixer=RX_mixer, cl_cost_function=cl_cost_function)
        QAOAinstance.set_init_function(init_function=init_state)
        theNiceQAOA = QAOAinstance.run(qarg=qarg, depth=5)

        import itertools
        def testCostFun(state):
            obj = 0
                #ljust do a minus 1 op if state is equiv to a given condition clause
            for index in range(len(clauses)):
                clause = clauses[index]
                sum = 0
                for index_aj in range(0,2):
                    sum += int(state[clause[index_aj]])
                if sum == clause[3] % 2:
                    obj -= 1 
            return obj
        
        
        #maxfive = sorted(theNiceQAOA, key=theNiceQAOA.get, reverse=True)[:1]
        for name in theNiceQAOA.keys():  # for name, age in dictionary.iteritems():  (for Python 2.x)
            if testCostFun(name) < 0:
                temp = False

                for clause in clauses:
                    sum = 0
                    for index_aj in range(0,2):
                        sum += int(name[clause[index_aj]])
                    if sum == clause[3] % 2:
                        temp  = True
                assert temp 
    



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.qaoa import QAOAProblem
from qrisp.qaoa.problems.maxIndepSetInfrastr import maxIndepSetCostOp, maxIndepSetclCostfct,  init_state
from qrisp.qaoa.mixers import RX_mixer
from qrisp import QuantumVariable
from qrisp.qaoa import approximation_ratio

import networkx as nx
import matplotlib.pyplot as plt
import numpy as np

def test_QAOAmaxIndepSet():
    # start with a graph G - from np adjacency matrix

    giraf = nx.erdos_renyi_graph(9,0.2)
    #print(nx.maximum_independent_set(giraf))
    nx_sol = nx.maximal_independent_set(giraf)
    print(nx_sol)
    #draw graph


    qarg = QuantumVariable(giraf.number_of_nodes())


    #Instanciate QAOA
    #mixer gets Graph as argument
    #cost function gets graph as argument 
    QAOAinstance = QAOAProblem(cost_operator=RX_mixer, mixer=maxIndepSetCostOp(giraf), cl_cost_function=maxIndepSetclCostfct(giraf))
    QAOAinstance.set_init_function(init_function=init_state)
    theNiceQAOA = QAOAinstance.run(qarg=qarg,depth=5)



    import itertools
    def aClcostFct(state, G):
        
        # we assume solution is right
        temp = True
        energy = 0 
        intlist = [s for s in range(len(list(state))) if list(state)[s] == "1"]
        # get all combinations of vertices in graph that are marked as |1> by the solution 
        combinations = list(itertools.combinations(intlist, 2))
        # if any combination is found in the list of G.edges(), the solution is wrong, and energy == 0
        for combination in combinations:
            if combination in G.edges():
                temp = False
        # else we just add the number of marked as |1> nodes
        if temp: 
            energy = -len(intlist)
        return(energy)

    
    # write correct sol test --> this done in the cost function... 
    for state in theNiceQAOA.keys():
        if aClcostFct(state=state, G=giraf) < 0:
            intlist = [s for s in range(len(list(state))) if list(state)[s] == "1"]
            # get all combinations of vertices in graph that are marked as |1> by the solution 
            combinations = list(itertools.combinations(intlist, 2))
            for combination in combinations:
                assert combination not in giraf.edges()


    #print the ideal solutions
    #print("5 most likely Solutions") 
    maxfive = sorted(theNiceQAOA, key=theNiceQAOA.get, reverse=True)[:5]
    for name in theNiceQAOA.keys():  # for name, age in dictionary.iteritems():  (for Python 2.x)
        if name in maxfive:
            print(name)
            print(aClcostFct(name, giraf))

    # write approximation ratio test
    metric_dict ={}
    metric_dict["counts"] = theNiceQAOA
    metric_dict["cost_func"] = maxIndepSetclCostfct(giraf)

    this = ["1" if s in nx_sol else "0" for s in range(giraf.number_of_nodes()) ]
    this2 = "".join(this)
    metric_dict["optimal_sol"] = this2
    
    assert approximation_ratio(metric_dict["counts"], metric_dict["optimal_sol"], maxIndepSetclCostfct(giraf))>= 0.1

    # test for full solution...
    integ = np.random.randint(1,7)
    integ2 = np.random.randint(0,1)
    integ2 = 0
    giraf1 = nx.erdos_renyi_graph(integ,integ2)
    qarg2 = QuantumVariable(giraf1.number_of_nodes())
    QAOAinstance2 = QAOAProblem(cost_operator= RX_mixer, mixer= maxIndepSetCostOp(giraf1), cl_cost_function=maxIndepSetclCostfct(giraf1)) 
    QAOAinstance2.set_init_function(init_function=init_state)
    theNiceQAOA2 = QAOAinstance2.run(qarg=qarg2, depth= 5)
    maxOne = sorted(theNiceQAOA2, key=theNiceQAOA2.get, reverse=True)[:1]
    if integ2 == 0:
        testStr = [str(1-integ2) * giraf1.number_of_nodes()]
    if integ2 == 1:
        testStr = [str(1-integ2) * giraf1.number_of_nodes()]
        string = "1"
        string += str(1-integ2) * (giraf1.number_of_nodes()-1)
        testStr = [string]
    #print(testStr)
    #print(maxOne)
    assert testStr == maxOne



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp import rz, rzz, x
import numpy as np
import copy
import networkx as nx
from qrisp.qiro.qiroproblems.qiro_utils import * 




def create_maxClique_replacement_routine( res, Graph, solutions, exclusions):
    """
    Creates a replacement routine for the problem structure, i.e. defines the replacement rules. 
    See the original paper for description of the update rules

    Parameters
    ----------
    res : dict
        Result dictionary of initial QAOA optimization procedure.
    Graph : nx.Graph
        The Graph defining the problem instance.
    solutions : List
        Qubits which have been found to be positive correlated, i.e. part of the problem solution.
    exclusions : List
        Qubits which have been found to be negative correlated, i.e. not part of the problem solution, or contradict solution qubits in accordance to the update rules.  

    Returns
    -------
    newGraph : nx.Graph
        Updated graph of the problem instance.
    solutions : List
        Updated set of solutions to the problem.
    sign : Int
        The sign of the correlation.
    exclusions : List
        Updated set of exclusions to the problem.
        
    """

    orig_edges = [list(item) for item in Graph.edges()]
    orig_nodes = list(Graph.nodes())

    #get the max_edge and eval the sum and sign
    max_item, sign = find_max(orig_nodes, orig_edges , res, solutions)

    newGraph = copy.deepcopy(Graph)

    # we just directly remove vertices from the graph 
    if isinstance(max_item, int):
        if sign > 0:
            border = list(Graph.adj[max_item].keys())
            border.append(max_item)
            to_remove = [int(item) for item in Graph.nodes() if item not in border]
            newGraph.remove_nodes_from( [item for item in Graph.nodes() if item not in border])
            solutions.append(max_item)
            exclusions += to_remove

        elif sign < 0:
            #remove item
            newGraph.remove_node(max_item)
            exclusions.append(max_item)

    else:
        if sign > 0:
            #keep the two items in solution and remove all that are not adjacent to both
            
            intersect = list(set(list(Graph.adj[max_item[0]].keys())) & set(list(Graph.adj[max_item[0]].keys())))
            intersect.append(max_item[0])
            intersect.append(max_item[1])
            to_remove = [int(item) for item in Graph.nodes() if item not in intersect]
            newGraph.remove_nodes_from([item for item in Graph.nodes() if item not in intersect])
            solutions.append(max_item[0])
            solutions.append(max_item[1])
            exclusions += to_remove

        elif sign < 0:
            #remove all that do not border on either! node
            union = list(Graph.adj[max_item[0]].keys())
            union += list(Graph.adj[max_item[1]].keys())
            union.append(max_item[0])
            union.append(max_item[1])
            to_remove = [int(item) for item in Graph.nodes() if item not in union]
            #to_delete = [item for item in Graph.nodes() if item not in union]
            newGraph.remove_nodes_from([item for item in Graph.nodes() if item not in union])
            exclusions += to_remove

    return newGraph, solutions, sign, exclusions



def create_maxClique_cost_operator_reduced(Graph, solutions = []):
    """
    |  Based on PennyLane unconstrained mixer implementation.
    |  Initial state in :math:`(|0>+|1>)^{\otimes n}` . 
    |  This operator is then adjusted to consider qubits that have been found to be a part of the problem solution.

    Parameters
    ----------
    G : nx.Graph
        Graph of the problem instance
    solutions : List
        Qubits which have been found to be positive correlated, i.e. part of the problem solution.
    
    Returns
    -------
    partialCostMixer : function
        The Operator to be applied to the problem ``QuantumVariable``

    """
    G_compl = nx.complement(Graph)
    def partialcostMixer(qv, gamma):
        for pair in list(G_compl.edges()):
            rzz(3*gamma, qv[pair[0]], qv[pair[1]])
            rz(-gamma, qv[pair[0]])
            rz(-gamma, qv[pair[1]])
        for i in Graph.nodes():
            if not i in solutions:
                rz(gamma, qv[i])


    return partialcostMixer









"""
BM: 18 nodes, seed 99
0.5933101570400762
normal
0.08909071112167886


"""


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.qaoa import QAOAProblem
from qrisp.qaoa.problems.maxCut import create_maxcut_cl_cost_function,create_maxcut_cost_operator
from qrisp.qaoa import RX_mixer
from qrisp import QuantumArray, QuantumVariable
import networkx as nx
from operator import itemgetter

def test_maxcut():
    def maxcut_obj(x,G):
        cut = 0
        for i, j in G.edges():
            if x[i] != x[j]:                        
                cut -= 1    
        return cut

    depth = 5

    from qrisp.default_backend import def_backend

    qrisp_sim = def_backend
    qaoa_backend = qrisp_sim

    ###### Trivial case with 1 edge
    G1 = nx.Graph()
    G1.add_edge(0, 1)

    qarg = QuantumArray(qtype = QuantumVariable(1), shape = len(G1))

    maxcut_instance1 = QAOAProblem(create_maxcut_cost_operator(G1), RX_mixer, create_maxcut_cl_cost_function(G1))

    res1 = maxcut_instance1.run(qarg, depth, mes_kwargs={"backend" : qaoa_backend}, max_iter = 50)
    best_cut1, best_solution1 = min([(maxcut_obj(x,G1),x) for x in res1.keys()], key=itemgetter(0))

    res_str1 = list(res1.keys())[0]
    best_cut1, best_solution1 = (maxcut_obj(res_str1,G1),res_str1)
    assert -best_cut1 == 1

    ####### Less trivial case
    G4 = nx.Graph()
    G4.add_edges_from([[0,3],[0,4],[1,3],[1,4],[2,3],[2,4]])

    qarg = QuantumArray(qtype = QuantumVariable(1), shape = len(G4))

    maxcut_instance4 = QAOAProblem(create_maxcut_cost_operator(G4), RX_mixer, create_maxcut_cl_cost_function(G4))
    res4 = maxcut_instance4.run(qarg, depth, mes_kwargs={"backend" : qaoa_backend}, max_iter = 50)
    best_cut4, best_solution4 = min([(maxcut_obj(x,G4),x) for x in res4.keys()], key=itemgetter(0))

    res_str4 = list(res4.keys())[0]
    best_cut4, best_solution1 = (maxcut_obj(res_str4,G4),res_str4)
    assert -best_cut4 == 6



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp import x, cx, invert, redirect_qfunction, custom_control
from qrisp.arithmetic.adders.qcla.classical_quantum.cq_carry_path import cq_calc_carry
from qrisp.arithmetic.adders.qcla.classical_quantum.cq_sum_path import cq_sum_path, cq_sum_path_direct_uncomputation
from qrisp.arithmetic.adders.gidney import cq_gidney_adder
from qrisp.misc.utility import bin_rep
from qrisp.circuit import fast_append
from qrisp.environments import QuantumEnvironment
from qrisp.core.session_merging_tools import merge

verify_manual_uncomputations = np.zeros(1)


# This function performs the in-place addition 
# b += a
# based on the higher radix qcla
# The overall radix can be specified as an exponential of the form
# R = radix_base**radix_exponent
@custom_control
def cq_qcla(a, b, radix_base = 2, radix_exponent = 1, t_depth_reduction = True, ctrl = None):
    
    if isinstance(a, int):
        a = bin_rep(a%(2**len(b)), len(b))[::-1]
    
    if len(a) > len(b):
        raise Exception("Tried to add QuantumFloat of higher precision onto QuantumFloat of lower precision")
    
    R = radix_base**radix_exponent
    
    merge([a,b])
    with fast_append(3):
        # The case that a only has a single qubit is simple.
        if len(b) == 1:
            if a[0] == "1":
                if ctrl is None:
                    x(b[0])
                else:
                    cx(ctrl, b[0])
            return
        elif len(b) <= R:
            cq_gidney_adder(a, b, ctrl = ctrl)
            return
        
        # Calculate the carry
        # Executing within a QuantumEnvironemnt accelerates the uncomputation algorithm
        # because it doesn't have to consider the operations appended outside of this function
        with QuantumEnvironment():
            merge([b[0].qs()])
            c = cq_calc_carry(a, b, radix_base, radix_exponent, ctrl = ctrl)
        
        
            if t_depth_reduction:
                cq_sum_path_direct_uncomputation(a,b,c,R, ctrl = ctrl)
            else:
                cq_sum_path(a,b,c,R, ctrl = ctrl)
                
                # To uncompute the carry we use Drapers strategy
                # CARRY(A,B) = CARRY(A, NOT(A+B))
                # We therefore bitflip the sum
                for i in range(len(b)):
                    x(b[i])
                
                # Contrary to Draper's adder we don't need to uncompute every carry digit.
                # Because of the above equivalence, the carries agree on every digit, so especially
                # on the digits representing the output of the calc_carry function. We can therefore
                # uncompute using calc_carry (even with higher radix) by inverting calc_carry.
                
                with invert():
                    #We use the redirect_qfunction decorator to steer the function onto c
                    redirect_qfunction(cq_calc_carry)(a, b, radix_base, radix_exponent, target = c, ctrl = ctrl)
                
                # Flip the sum back
                for i in range(len(b)):
                    x(b[i])
            
        # Delete c
        c.delete(verify = bool(verify_manual_uncomputations[0]))
            
        
        


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.qaoa.qaoa_problem import QAOAProblem
import inspect
import numpy as np
import copy

class QIROProblem(QAOAProblem):
    r"""
    Structure to run QIRO algorithms. The idea is based on the paper by J. Finzgar et al. (https://arxiv.org/pdf/2308.13607.pdf).

    This class encapsulates the cost operator, mixer operator, and classical cost function for a specific QIRO problem instance. It also provides methods to set the initial state preparation function, classical cost post-processing function, and optimizer for the problem.

    For a quick demonstration, we import the relevant functions from already implemented problem instances:
        
    ::
        
        # imports 
        from qrisp.qaoa.qiro_problem import QIROProblem
        from qrisp.qaoa.qaoa_problem import QAOAProblem
        from qrisp.qaoa.problems.create_rdm_graph import create_rdm_graph
        from qrisp.qaoa.problems.maxCliqueInfrastr import maxCliqueCostfct, maxCliqueCostOp
        from qrisp.qaoa.qiroproblems.qiroMaxCliqueInfrastr import * 
        from qrisp.qaoa.mixers import RX_mixer
        from qrisp.qaoa.qiro_mixers import qiro_init_function, qiro_RXMixer
        from qrisp import QuantumVariable
        import matplotlib.pyplot as plt
        import networkx as nx



        # First we define a graph via the number of nodes and the QuantumVariable arguments
        num_nodes = 15
        G = create_rdm_graph(num_nodes,0.7, seed =  99)
        Gtwo = create_rdm_graph(num_nodes,0.7, seed =  99)
        qarg = QuantumVariable(G.number_of_nodes())
        qarg2 = QuantumVariable(Gtwo.number_of_nodes())

        # set simulator shots
        mes_kwargs = {
            #below should be 5k
            "shots" : 5000
            }

        #assign cost_function and maxclique_instance, normal QAOA
        testCostFun = maxCliqueCostfct(Gtwo)
        maxclique_instance = QAOAProblem(maxCliqueCostOp(G), RX_mixer, maxCliqueCostfct(G))

        # assign the correct new update functions for qiro from above imports
        qiro_instance = QIROProblem(problem = Gtwo,  
                                    replacement_routine = create_maxClique_replacement_routine, 
                                    cost_operator = create_maxClique_cost_operator_reduced,
                                    mixer = qiro_RXMixer,
                                    cl_cost_function = maxCliqueCostfct,
                                    init_function = qiro_init_function
                                    )


        # We run the qiro instance and get the results!
        res_qiro = qiro_instance.run_qiro(qarg=qarg, depth = 3, n_recursions = 2, 
                                        #mes_kwargs = mes_kwargs
                                        )
        # and also the final graph, that has been adjusted
        final_Graph = qiro_instance.problem

        # get the normal QAOA results for a comparison
        res_qaoa = maxclique_instance.run( qarg = qarg2, depth = 3)


        # We can then also print the top 5 results for each...
        print("QAOA 5 best results")
        maxfive = sorted(res_qaoa, key=res_qaoa.get, reverse=True)[:5]
        for key, val in res_qaoa.items(): 
            if key in maxfive:
                print(key)
                print(testCostFun({key:1}))


        print("QIRO 5 best results")
        maxfive = sorted(res_qiro, key=res_qiro.get, reverse=True)[:5]
        for key, val in res_qiro.items():  
            if key in maxfive:
                
                print(key)
                print(testCostFun({key:1}))

        # or compare it with the networkx result of the max_clique algorithm...
        print("Networkx solution")
        print(nx.approximation.max_clique(Gtwo))

        # and finally, we draw the final graph and the original graphs to compare them!
        plt.figure(1)
        nx.draw(final_Graph, with_labels = True)

        plt.figure(2)
        nx.draw(G, with_labels = True)
        plt.show()  
        

    For an in-depth tutorial, make sure to check out :ref:`QIROmaxClique`!
        
    Parameters
    ----------
    problem : Any
        The problem structure to be considered for the algorithm. For example in the case of MaxClique a graph, or MaxSat a list of clauses
    qaoaProblem : QAOAProblem
        A :ref:`QAOAProblem` instance to be used for establishing the correlations.
    replacement_routine  : function
        A routine for adjusting the problem after the highest correlation value has been found.
    qiro_cost_operator  : function
        New ``cost_operator`` for the :ref:`QAOAProblem` instance.
    qiro_mixer_operator  : function
        New ``mixer_operator`` for the :ref:`QAOAProblem` instance.
    qiro_init_function  : function
        New ``mixer_operator`` for the :ref:`QAOAProblem` instance.

    """
    
    #try without QAOAProblem
    """ def __init__(self, problem, qaoaProblem, replacement_routine , qiro_cost_operator, qiro_mixer, qiro_init_function, ):
        super().__init__(qaoaProblem.cost_operator, qaoaProblem.mixer, qaoaProblem.cl_cost_function)
        self.problem = problem
        self.replacement_routine = replacement_routine
        self.qiro_cost_operator = qiro_cost_operator
        self.qiro_mixer = qiro_mixer
        self.qiro_init_function = qiro_init_function """
    
    def __init__(self, problem, replacement_routine , cost_operator, mixer, cl_cost_function, init_function):
        super().__init__(cost_operator(problem), mixer(), cl_cost_function(problem))
        self.problem = problem
        self.replacement_routine = replacement_routine
        self.qiro_cost_operator = cost_operator
        self.qiro_mixer = mixer
        self.init_function = init_function()
        self.qiro_init_function = init_function
    
    def run_qiro(self, qarg, depth, n_recursions,  mes_kwargs = {}, max_iter = 50):
        """
        Run the specific QIRO problem instance with given quantum arguments, depth of QAOA circuit,
        measurement keyword arguments (mes_kwargs) and maximum iterations for optimization (max_iter).
        
        Parameters
        ----------
        qarg : QuantumVariable
            The quantum variable to which the QAOA circuit is applied.
        depth : int
            The depth of the QAOA circuit.
        n_recursions : int
            The number of QIRO replacement iterations.  
        mes_kwargs : dict, optional
            The keyword arguments for the measurement function. Default is an empty dictionary.
        max_iter : int, optional
            The maximum number of iterations for the optimization method. Default is 50.

        Returns
        -------
        opt_res : dict
            The optimal result after running QAOA problem for a specific problem instance. It contains the measurement results after applying the optimal QAOA circuit to the quantum variable.
        """       

        from qrisp import QuantumVariable
        res= QAOAProblem.run(self, qarg, depth,  mes_kwargs, max_iter)

        corr_vals = []
        solutions = []
        exclusions = []

        for index in range(n_recursions):
            
            new_problem, solutions , sign, exclusions = self.replacement_routine(res, self.problem, solutions , exclusions)
            
            corr_vals.append(sign)    
            self.problem = new_problem
            self.cost_operator = self.qiro_cost_operator(new_problem, solutions=solutions)
            self.mixer = self.qiro_mixer(solutions=solutions, exclusions = exclusions)
            self.init_function = self.qiro_init_function(#problem = new_problem, 
                                                         solutions=solutions, exclusions = exclusions)

            new_qarg = QuantumVariable(len(qarg))
            res = QAOAProblem.run(self, new_qarg, depth,  mes_kwargs, max_iter)

        return res











from qrisp import QuantumCircuit
from qrisp.core import QuantumVariable
from qrisp import ControlledOperation
from qrisp.interface import BackendClient
import numpy as np

def create_tket_instruction(op):
    from pytket import OpType
    
    from pytket.circuit import CircBox
    if op.name == "rxx":
        tket_ins = OpType.XXPhase
    elif op.name == "rzz":
        tket_ins = OpType.ZZPhase
    elif op.name == "ryy":
        tket_ins = OpType.YYPhase
    elif op.name == "measure":
        tket_ins = OpType.Measure
    elif op.name == "swap":
        tket_ins = OpType.SWAP
    elif op.name == "h":
        tket_ins = OpType.H
    elif op.name == "p":
        tket_ins =  OpType.Rz
    elif op.name == "x":
        tket_ins = OpType.X
    elif op.name == "y":
        tket_ins = OpType.Y
    elif op.name == "z":
        tket_ins = OpType.Z
    elif op.name == "rx":
        tket_ins = OpType.Rx
    elif op.name == "ry":
        tket_ins = OpType.Ry
    elif op.name == "rz":
        tket_ins = OpType.Rz
    elif op.name == "s":
        tket_ins = OpType.S
    elif op.name == "s_dg":
        tket_ins = OpType.Sdg
    elif op.name == "t":
        tket_ins = OpType.T
    elif op.name == "t_dg":
        tket_ins = OpType.Tdg
    elif op.name == "u3":
        tket_ins = OpType.U3
    
    elif op.definition:
        # if complex definition we create an abstract circBox for the section
        tket_definition = pytket_converter(op.definition, boxFlag= True)
        # might need adjustment
        #if len(op.definition.clbits):
        if tket_definition.n_qubits != op.num_qubits:
            raise Exception
        
        tket_ins = CircBox(tket_definition)

    else:
        raise Exception("Could not convert operation " + str(op.name) + " to PyTket") 

    return tket_ins



def pytket_converter(qc, boxFlag = False):
    from pytket import Circuit, Qubit, Bit, OpType
    
    from pytket.circuit import CircBox, QControlBox, Op

    # This dic gives the qiskit qubits/clbits when presented with their identifier
    qubit_dic = {}
    tket_qc = Circuit()
    #stringListQubs = []
    tketQubits = []
    for i in range(len(qc.qubits)):
        # add a named qubit
        tketQubits.append(Qubit(name = str(qc.qubits[i].identifier), index = i))
        qubit_dic[qc.qubits[i].identifier] = tketQubits[-1]
        tket_qc.add_qubit(tketQubits[-1])
    
    # Flag for alternative qubit assignment if we try to create an abstract CircBox
    if boxFlag:
        tket_qc = Circuit(len(qc.qubits))
        qubit_dic = dict()
        for i in range(len(qc.qubits)):
            qubit_dic[qc.qubits[i].identifier] = i

    clbit_dic = {}
    # Add Clbits
    tketClbits = []
    if len(qc.clbits):
        c_reg = tket_qc.add_c_register(name = "creg_std", size = len(qc.clbits))
    for i in range(len(qc.clbits)):
        clbit_dic[qc.clbits[i].identifier] = c_reg[i]
        #this will hopefully be used one day, when other simulators other than Aer are used with this backend, or... quantinuum decides to fix their backend integration
        # will throw an error on Aer backend and QASM converter, since they apparently only supports a single classical register, which is a lie
        """ tketClbits.append(Bit( name = str(qc.clbits[i].identifier)))
        clbit_dic[qc.clbits[i].identifier] = tketClbits[-1]
        tket_qc.add_bit(tketClbits[-1]) """

    for i in range(len(qc.data)):
        op = qc.data[i].op

        params = list(op.params)
        # Prepare qubits
        qubit_list = [qubit_dic[qubit.identifier] for qubit in qc.data[i].qubits]
        clbit_list = [clbit_dic[clbit.identifier] for clbit in qc.data[i].clbits]

        if op.name in ["cp", "p", "rx", "rz", "ry", "rxx", "rzz", "ryy", "u1", "u3"]: #and not boxFlag:
            #pytket expects angles in pi multiples
            params = [index/np.pi for index in params]  

        #add_gate
        if op.name in ["qb_alloc", "qb_dealloc"]:
            continue

        elif op.name == "cx":
            # maybe adjustment necessary here
            if hasattr(op, "ctrl_state"):
                tket_ins = OpType.CX
            else:
                tket_ins = OpType.CX

        elif op.name == "cy":
            if hasattr(op, "ctrl_state"):
                tket_ins = OpType.CY
            else:
                tket_ins = OpType.CY

        elif op.name == "cz":
            if hasattr(op, "ctrl_state"):
                tket_ins = OpType.CZ
            else:
                tket_ins = OpType.CZ
        
        elif op.name == "cp":
            if hasattr(op, "ctrl_state"):
                tket_ins = OpType.CRz
            else:
                tket_ins = OpType.Rz
        
        elif op.name == "sx":
            #bugged -> params empty
            params = []
            tket_ins = OpType.SX
        elif op.name == "sx_dg":
            #bugged -> params empty
            params = []
            tket_ins = OpType.SXdg

        elif op.name == "u1":
            params[0] = params[0]/np.pi
            #bugged
            tket_ins = OpType.Rz    
        elif op.name == "id":
            params = []
            #bugged
            tket_ins = OpType.noop


        elif issubclass(op.__class__, ControlledOperation):
            base_name = op.base_operation.name

            if len(base_name) == 1:
                base_name = base_name.upper()

            if 0==1:
            #if op.base_operation.definition:
                # old code relic -- buggy 
                # base_operation.definition doesnt convert correctly if mutiple abstract/costum gates are included
                tket_definition = pytket_converter(op.base_operation.definition)
                base_gate = tket_definition
                if isinstance( base_gate, Circuit):
                    tket_definition = pytket_converter(op.base_operation.definition, boxFlag=True)
                    tket_definition.name = base_name
                    tket_ins = CircBox(tket_definition)
                else: 
                    tket_ins = QControlBox(Op.create(base_gate) , len(op.controls))


            else:
                base_gate = pytket_converter(op.definition)
                
                if isinstance( base_gate, Circuit):
                    #circuit is returned as abstract CircBox --> pytket specific
                    tket_definition = pytket_converter(op.definition, boxFlag=True)
                    tket_definition.name = base_name
                    tket_ins = CircBox(tket_definition)

                else: 
                    #else a simpler multi controlled gate can be created
                    tket_ins = QControlBox(Op.create(base_gate) , len(op.controls))
                

        else:
            tket_ins = create_tket_instruction(op)

        if isinstance(tket_ins, CircBox):
            tket_qc.add_circbox(tket_ins, qubit_list)

        elif isinstance(tket_ins, QControlBox):
            tket_qc.add_qcontrolbox(tket_ins, qubit_list)
            
        elif len(clbit_list):
            # add other isinstance checks from above here aswell?
            tket_qc.add_gate(tket_ins, params, qubit_list + clbit_list)

        else:
            tket_qc.add_gate(tket_ins, params, qubit_list)

    return tket_qc
    













"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp import QuantumVariable, h, barrier, rz, rx , cx, QuantumArray,x 
import numpy as np
from scipy.optimize import minimize
from sympy import Symbol
import itertools
from _collections_abc import Iterable





def maxSatCostOp(problem):

    """
    |  Implementation for MaxSat Cost-Operator, in accordance to the original QAOA-paper.
    |  For each clause :math:`C_i` apply :math:`C_i(x) = 1 - \prod_j x_j` (replace :math:`x_j` with 
    :math:`(1-x_j)` if :math:`x_j` is negated). The problem operator is acquired with the substitution 
    :math:`x_j = (I - Z_j)` , where :math:`Z_j` is the Pauli- :math:`Z` operator applied to the :math:`j` -th qubit.
    
    Parameters
    ----------
    clauses : List(Lists)
        Clauses to be considered for MaxSat. Should be given as a list of lists and 1-indexed instead 0-indexed, see example

        
    Returns
    -------
    QuantumCircuit: qrisp.QuantumCircuit
        the Operator applied to the circuit-QuantumVariable

    Examples
    --------
    >>> clauses11 = (6, [[1,2,-3], [1,4,-6], [4,5,6], [1,3,-4], [2,4,5], [1,3,5], [-2,-3,6]])
    
    |  Explanation: 
    |  First entry of tuple is the number of variables, second is the clauses
    |  Clause [1, 2, -4] is fulfilled by the QuantumStates "1100", "1110" 

    * if the sign is positive : the index has of the QuantumVariable to be "1", if negative sign the index has to be "0".
    * indices not mentioned can be "0" or "1" (the third integer in this case).
    * start with 1 in your clauses! because ``int 0`` has no sign and this is problematic. 
    * we want to keep the mathematical formulation, so no handing over strings!

    Assign the operators

    >>> cost_operator=maxSatCostOp(clauses11),
    >>> mixer=RX_mixer 

    
    """
    

    clauses = problem[1]

    if not isinstance(clauses, Iterable):
        raise Exception("Wrong structure of problem - clauses have to be iterable!")
    satlen = len(clauses[0])
    for index in range(len(clauses)):
        if not satlen ==len(clauses[index]):
            raise Exception("Wrong structure of problem - inconsistent length in clause " + str(index))
        if not isinstance(clauses[index], Iterable):
            raise Exception("Wrong structure of problem - each clauses has to be an array! Problem in clause " + str(index))
        for item in clauses[index]:
            if not isinstance(item, int):
                raise Exception("Wrong structure of problem - each literal has to an int!")



    def maxSatcostEmbed(qv , gamma):
        #clauses = clauses
        
        qc = qv

        for numClause in range(len(clauses)):
            #go through clauses
            clause = clauses[numClause]
        
            for lenofCombination in range(1,satlen+1):
                # get all combinations to assign rz, rzz, rzzz, rzzzzzz....
                combinations = list(itertools.combinations(clause, lenofCombination))
                #print(combinations)
                
                #len == 1: just rz
                if lenofCombination == 1:
                    for index in range(len(combinations)):
                        # sign for rz mixer
                        signu = - np.sign(combinations[index][0])
                        # always have the "-1" at the end since clauses are not initiated with 0, see above
                        #print(abs( combinations[index][0] - 1 ))
                        rz(signu *  gamma/8, qc[ abs( combinations[index][0]) -1 ] )

                else:
                    #for all combinations of this length 
                    for index in range(len(combinations)):
                        signu = 1

                        # up to final value in combination perform cx gates --> set up the rzz, rzzz, rzzzz ... gates
                        for index2 in range(lenofCombination-1):
                            # (also remember the sign)
                            signu *= - np.sign(combinations[index][index2])
                            cx(qc[abs( combinations[index][index2] ) -1], qc[abs( combinations[index][index2+1] ) -1])

                        signu *= np.sign(combinations[index][lenofCombination-1])
                        # finalize rzz, rzzz, rzzzz ... gates
                        rz(signu*gamma/8, qc[abs( combinations[index][lenofCombination-1] ) -1])
                        
                        # and cnot gates back 
                        for index2 in reversed(range(lenofCombination-1)):
                            cx(  qc[abs(combinations[index][index2] ) -1], qc[abs(combinations[index][index2+1] ) -1 ]) 

        return qc
    
    return maxSatcostEmbed




def clausesdecoder(problem): 
    """
    Decodes the clause arrays to represent binary bitstrings, that fulfill the clauses
    --> used to calculate objective function, i.e. the classical cost function

    
    Parameters
    ----------
    clauses : List(List)
        clauses to be considered for maxSat (! 1-indexed, instead of 0-indexed, see example above)
    numVars :  int
        the total amount of Variables considered within the clauses

    Returns
    -------
    decodedClauses : Iterable(tuple)
        clauses as bitstrings, to be compared to the QAOA-bitstring results

    Examples
    --------
    As above:

    >>> clauses11 = [[1,2,-3], [1,4,-6], [4,5,6], [1,3,-4], [2,4,5], [1,3,5], [-2,-3,6]]
    >>> decodedClauses = clausesdecoder( clauses = clauses11, numVars = 6)

    Assign ``cost_function``
    
    >>> cl_cost_function=maxSatclCostfct(decodedClauses)

    

    """

    numVars = problem[0]
    clauses = problem[1]

    # create all bitstring possibilites
    binStrings = list(itertools.product([0,1], repeat=numVars))
    decodedClauses = []
    for indexClauses in range(len(clauses)):
        # create temp placeholder 
        tempIter = binStrings 
        clause = clauses[indexClauses]
        for index in clause:
            val = abs(index)-1 
            temp = 1 if np.sign(index) == 1 else 0 
            # placeholder only keeps to bitstrings that fulfill the clause-conditions
            tempIter = [ thing for thing in tempIter if thing[val] == temp ]
        # turn to str, since QAOA return is dict of str
        tempIter2 = ["".join(map(str, item))  for item in tempIter]
        decodedClauses.append(tuple(tempIter2))

    return decodedClauses



def maxSatclCostfct(problem):

    """

    Parameters
    ----------
    decodedClauses : Iterable(tuple)
        clauses as bitstrings, as returned by ``clausesdecoder()`` helper function


    Returns
    -------
    Costfunction : function
        the classical function for the problem instance, which takes a dictionary of measurement results as input

    """


    decodedClauses = clausesdecoder(problem)

    def setupaClCostfct(res_dic):
        energy = 0
        total_counts = 0
        for state in list(res_dic.keys()):
            obj = 0
            #literally just do a minus 1 op if state is equiv to a given condition
            for index in range(len(decodedClauses)):
                if state in decodedClauses[index]:
                    obj -= 1 
            energy += obj * res_dic[state]
            total_counts += res_dic[state]
        #print(energy/total_counts)

        return energy/total_counts
    
    return setupaClCostfct


def init_state(qv):
    h(qv)
    return qv




"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp import QuantumFloat, QuantumVariable, QuantumBool, qq_calc_carry, cq_calc_carry, h, multi_measurement, qcla, bin_rep, control, x


def test_qq_carry_generation():
    from qrisp.uncomputation import verify
    verify[0] = 1
    from qrisp.arithmetic.adders.qcla.quantum_quantum.qq_qcla_adder import verify_manual_uncomputations
    verify_manual_uncomputations[0] = 1
    
    
    # Test functions
    # Function to classically calculate the carry value
    def full_adder_carry(bitstring1, bitstring2):
        if len(bitstring1) != len(bitstring2):
            raise ValueError("Bitstrings must be of equal length")

        n = len(bitstring1)
        carry = 0
        carry_values = []

        for i in range(n):
            a = int(bitstring1[i])
            b = int(bitstring2[i])

            # Calculate the carry-out (Cout) for this bit
            sum_ab = a + b + carry
            carry_out = sum_ab // 2

            carry_values.append(carry_out)

            # Update the carry for the next iteration
            carry = carry_out

        return ''.join(str(c) for c in carry_values)

        
    n = 4
    for n in range(3, 8):
        for r in range(2,n):
            a = QuantumVariable(n)
            b = QuantumVariable(n)
            
            h(a)
            h(b)
                    
            c = qq_calc_carry(a, b, r)
            
            mes_res = multi_measurement([a, b, c])
            
            for k in mes_res.keys():
                if k[2] != full_adder_carry(k[0], k[1])[:-1]:
                    print(k[0])
                    print(k[1])
                    print(k[2])
                    print(full_adder_carry(k[0], k[1])[:-1])
                    
                    assert False
                    
    from qrisp.uncomputation import verify
    verify[0] = 0
    from qrisp.arithmetic.adders.qcla.quantum_quantum.qq_qcla_adder import verify_manual_uncomputations
    verify_manual_uncomputations[0] = 0

def test_qq_qcla_adder():
    
    from qrisp.uncomputation import verify
    verify[0] = 1
    from qrisp.arithmetic.adders.qcla.quantum_quantum.qq_qcla_adder import verify_manual_uncomputations
    verify_manual_uncomputations[0] = 1
    
    import time
    
    t0 = time.time()
    for radix_base in [2,3]:
        for radix_exponent in [1,0,2]:
            for m in range(1, 7):        
                for n in range(1, m):
                    
                    a = QuantumFloat(n)
                    b = QuantumFloat(m)
                    c = QuantumFloat(m)
                    
                    h(a)
                    h(b)
                    c[:] = b
                    
                    qcla(a, b, radix_base = radix_base, radix_exponent = radix_exponent)
                    mes_res = multi_measurement([a, c, b])
                    
                    for k in mes_res.keys():
                        if k[2] != (k[0] + k[1])%2**m:
                            print(k[0])
                            print(k[1])
                            print(k[2])
                            print((k[0]+k[1])%2**m)
                            
                            assert False
                    
                    statevector_arr = a.qs.compile().statevector_array()
                    angles = np.angle(
                        statevector_arr[
                            np.abs(statevector_arr) > 1 / 2 ** ((a.size + b.size) / 2 + 1)
                        ]
                    )
        
                    # Test correct phase behavior
                    assert np.sum(np.abs(angles)) < 0.1
                    
                    
    print(time.time()-t0)
    from qrisp.uncomputation import verify
    verify[0] = 0
    from qrisp.arithmetic.adders.qcla.quantum_quantum.qq_qcla_adder import verify_manual_uncomputations
    verify_manual_uncomputations[0] = 0
    
    a = QuantumFloat(8)
    b = QuantumFloat(8)
    a[:] = 4
    b[:] = 15
    qcla(a, b)

    from qrisp import t_depth_indicator
    gate_speed = lambda x : t_depth_indicator(x, epsilon = 2**-10)
    qc = b.qs.compile(gate_speed = gate_speed, compile_mcm = True)
    assert qc.t_depth() < 20

    qc = b.qs.compile(workspace = 10, gate_speed = gate_speed, compile_mcm = True)
    assert qc.t_depth() < 9

    a = QuantumFloat(40)
    b = QuantumFloat(40)
    x(a)
    x(b)
    qcla(a, b)
    qc = b.qs.compile(workspace = 50, gate_speed = gate_speed, compile_mcm = True)
    assert qc.t_depth() < 22


def test_cq_carry_generation():
    from qrisp.uncomputation import verify
    verify[0] = 1
    from qrisp.arithmetic.adders.qcla.classical_quantum.cq_qcla_adder import verify_manual_uncomputations
    verify_manual_uncomputations[0] = 1
    
    # Test functions
    # Function to classically calculate the carry value
    def full_adder_carry(bitstring1, bitstring2):
        if len(bitstring1) != len(bitstring2):
            raise ValueError("Bitstrings must be of equal length")

        n = len(bitstring1)
        carry = 0
        carry_values = []

        for i in range(n):
            a = int(bitstring1[i])
            b = int(bitstring2[i])

            # Calculate the carry-out (Cout) for this bit
            sum_ab = a + b + carry
            carry_out = sum_ab // 2

            carry_values.append(carry_out)

            # Update the carry for the next iteration
            carry = carry_out

        return ''.join(str(c) for c in carry_values)

        
    n = 4
    for n in [3, 4, 7]:
        for r in range(2,n):
            for a in range(2**n):
                b = QuantumVariable(n)
                ctrl = QuantumBool()
                
                h(ctrl)
                h(b)
                        
                c = cq_calc_carry(a, b, r, ctrl = ctrl[0])
                
                mes_res = multi_measurement([b, c, ctrl])
                
                for k in mes_res.keys():
                    
                    if k[2]:
                        if k[1] != full_adder_carry(k[0], bin_rep(a, n)[::-1])[:-1]:
                            print(k[0])
                            print(k[1])
                            print(k[2])
                            print(bin_rep(a, n)[::-1])
                            print(full_adder_carry(k[0], bin_rep(a, n)[::-1])[:-1])
                            
                            assert False
                    else:
                        assert k[1] == (n-1)*"0"
                        
                b = QuantumVariable(n)
                
                h(b)
                        
                c = cq_calc_carry(a, b, r)
                
                mes_res = multi_measurement([b, c])
                
                for k in mes_res.keys():
                    if k[1] != full_adder_carry(k[0], bin_rep(a, n)[::-1])[:-1]:
                        print(k[0])
                        print(k[1])
                        print(k[2])
                        print(full_adder_carry(k[0], k[1])[:-1])
                        
                        assert False
                    
    from qrisp.uncomputation import verify
    verify[0] = 0
    from qrisp.arithmetic.adders.qcla.classical_quantum.cq_qcla_adder import verify_manual_uncomputations
    verify_manual_uncomputations[0] = 0
    


def test_cq_qcla_adder():
    
    from qrisp.uncomputation import verify
    verify[0] = 1
    from qrisp.arithmetic.adders.qcla.classical_quantum.cq_qcla_adder import verify_manual_uncomputations
    verify_manual_uncomputations[0] = 1
    
    import time
    
    t0 = time.time()
    for radix_base in [2,3]:
        for radix_exponent in [1,0,2]:
            for m in range(1, 6):        
                for a in range(2**m):
                    
                    b = QuantumFloat(m)
                    c = QuantumFloat(m)
                    ctrl = QuantumBool()
                    
                    h(ctrl)
                    h(b)
                    c[:] = b
                    
                    qcla(a, b, radix_base = radix_base, radix_exponent = radix_exponent, ctrl = ctrl[0])
                    mes_res = multi_measurement([c, b, ctrl])

                    statevector_arr = b.qs.compile().statevector_array()
                    angles = np.angle(
                        statevector_arr[
                            np.abs(statevector_arr) > 1 / 2 ** ((c.size + b.size) / 2 + 1)
                        ]
                    )
        
                    # Test correct phase behavior
                    assert np.sum(np.abs(angles)) < 0.1

                    for c, b, ctrl in mes_res.keys():
                        
                        if ctrl:
                            if b != (c + a)%2**m:
                                print(b)
                                print(c)
                                print(a)
                                assert False
                        else:
                            assert b == c
                            
                    
                    # Test non-controlled version
                    
                    b = QuantumFloat(m)
                    c = QuantumFloat(m)
                    
                    h(b)
                    c[:] = b

                    qcla(a, b, radix_base = radix_base, radix_exponent = radix_exponent)
                    mes_res = multi_measurement([c, b])

                    statevector_arr = b.qs.compile().statevector_array()
                    angles = np.angle(
                        statevector_arr[
                            np.abs(statevector_arr) > 1 / 2 ** ((c.size + b.size) / 2 + 1)
                        ]
                    )
        
                    # Test correct phase behavior
                    assert np.sum(np.abs(angles)) < 0.1
                    
                    
                    
                    for c, b in mes_res.keys():
                        if b != (c + a)%2**m:
                            print(b)
                            print(c)
                            print(a)
                            assert False
                    
                    
    from qrisp.uncomputation import verify
    verify[0] = 0
    from qrisp.arithmetic.adders.qcla.classical_quantum.cq_qcla_adder import verify_manual_uncomputations
    verify_manual_uncomputations[0] = 0
    
    b = QuantumFloat(10)
    b += 4
    qbl = QuantumBool()
    h(qbl)

    with control(qbl):
        
        qcla(16, b)
        
    assert multi_measurement([qbl, b]) == {(True, 20) : 0.5, (False, 4) : 0.5}
                    
    b = QuantumFloat(8)
    a = 4
    b[:] = 15
    qcla(a, b)
    
    from qrisp import t_depth_indicator
    gate_speed = lambda x : t_depth_indicator(x, epsilon = 2**-10)
    qc = b.qs.compile(gate_speed = gate_speed, compile_mcm = True)
    assert qc.t_depth() < 11
    
    qc = b.qs.compile(workspace = 10, gate_speed = gate_speed, compile_mcm = True)
    assert qc.t_depth() < 7
    
    a = QuantumFloat(40)
    b = QuantumFloat(40)
    qcla(a, b)
    qc = b.qs.compile(workspace = 50, gate_speed = gate_speed, compile_mcm = True)
    assert qc.t_depth() < 21



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.interface import BackendClient


class VirtualBackend(BackendClient):
    """
    This class provides a virtual backend for circuit execution.
    Virtual means that the server is running on the same machine as a separate
    Python thread. This structure allows setting up convenient wrappers for foreign/
    modified circuit dispatching code.

    Circuits can be run using ``virtual_backend_instance.run(qc)``.
    The function that should be used to run a circuit can be specified during
    construction using the ``run_func`` parameter.


    Parameters
    ----------
    run_func : function
        A function that recieves a QuantumCircuit, an integer specifiying the amount of
        shots and a token in the form of a string. It returns the counts as a dictionary
        of bitstrings.
    name : str, optional
        A name for the virtual backend. The default is None.
    port : int, optional
        The port on which to listen. The default is None.

    Examples
    --------


    We set up a VirtualBackend, which prints the received QuantumCircuit and returns the
    results of the QASM simulator. ::


        def run_func(qasm_str, shots, token = ""):
            
            from qiskit import QuantumCircuit
            
            qiskit_qc = QuantumCircuit.from_qasm_str(qasm_str)

            print(qiskit_qc)

            from qiskit import Aer
            qiskit_backend = Aer.get_backend('qasm_simulator')

            #Run Circuit on the Qiskit backend
            return qiskit_backend.run(qiskit_qc, shots = shots).result().get_counts()


    >>> from qrisp.interface import VirtualBackend
    >>> example_backend = VirtualBackend(run_func)
    >>> from qrisp import QuantumFloat
    >>> qf = QuantumFloat(3)
    >>> qf[:] = 4
    >>> qf.get_measurement(backend = example_backend)
                 ┌─┐
    qf.0:   ─────┤M├──────
                 └╥┘┌─┐
    qf.1:   ──────╫─┤M├───
            ┌───┐ ║ └╥┘┌─┐
    qf.2:   ┤ X ├─╫──╫─┤M├
            └───┘ ║  ║ └╥┘
    cb_0: 1/══════╩══╬══╬═
                  0  ║  ║
    cb_1: 1/═════════╩══╬═
                     0  ║
    cb_2: 1/════════════╩═
    {4: 1.0}

    """

    def __init__(self, run_func, port=None):

        from qrisp.interface import BackendServer

        # Create BackendServer
        self.backend_server = BackendServer(
            run_func, "localhost", port=port
        )
        # Run the server (runs in the background)
        self.backend_server.start()
        # Connect client

        super().__init__(api_endpoint="localhost", port=port)

    def run(self, qc, shots, token=""):
        """
        Executes the function run_func specified at object creation.

        Parameters
        ----------
        qc : QuantumCircuit
            The QuantumCircuit to run.
        shots : int
            The amount of shots to perform.

        Returns
        -------
        res : dict
            A dictionary containing the measurement results.

        """
        return super().run(qc, shots)


class VirtualQiskitBackend(VirtualBackend):
    """
    This class instantiates a VirtualBackend using a Qiskit backend.
    This allows easy access to Qiskit backends through the qrisp interface.


    Parameters
    ----------
    backend : Qiskit backend object, optional
        A Qiskit backend object, which runs QuantumCircuits. The default is
        Aer.get_backend('qasm_simulator').
    port : int, optional
        The port to listen. The default is 8079.

    Examples
    --------

    We evaluate a QuantumFloat multiplication on the QASM-simulator.

    >>> from qrisp import QuantumFloat
    >>> from qrisp.interface import VirtualQiskitBackend
    >>> from qiskit import Aer
    >>> example_backend = VirtualQiskitBackend(backend = Aer.get_backend('qasm_simulator'))
    >>> qf = QuantumFloat(4)
    >>> qf[:] = 3
    >>> res = qf*qf
    >>> res.get_measurement(backend = example_backend)
    {9: 1.0}


    """

    def __init__(self, backend=None, port=9011):
        if backend is None:
            
            try:
                from qiskit import Aer
                backend = Aer.get_backend("qasm_simulator")
            except ImportError:
                from qiskit.providers.basic_provider import BasicProvider
                backend = BasicProvider().get_backend('basic_simulator')

        # Create the run method
        def run(qasm_str, shots, token = ""):
            # Convert to qiskit
            from qiskit import QuantumCircuit

            qiskit_qc = QuantumCircuit.from_qasm_str(qasm_str)

            
            #Make circuit with one monolithic register
            new_qiskit_qc = QuantumCircuit(len(qiskit_qc.qubits), len(qiskit_qc.clbits))
            for instr in qiskit_qc:
                new_qiskit_qc.append(instr.operation, 
                                     [qiskit_qc.qubits.index(qb) for qb in instr.qubits],
                                     [qiskit_qc.clbits.index(cb) for cb in instr.clbits])

            from qiskit import transpile
            qiskit_qc = transpile(new_qiskit_qc, backend=backend)
            # Run Circuit on the Qiskit backend
            qiskit_result = backend.run(qiskit_qc, shots=shots).result().get_counts()
            # Remove the spaces in the qiskit result keys
            result_dic = {}
            import re

            for key in qiskit_result.keys():
                counts_string = re.sub(r"\W", "", key)
                result_dic[counts_string] = qiskit_result[key]

            return result_dic

        # Call VirtualBackend constructor
        if isinstance(backend.name, str):
            name = backend.name
        else:
            name = backend.name()

        super().__init__(run, port=port)


class QiskitRuntimeBackend(VirtualBackend):
    """
    This class instantiates a VirtualBackend using a Qiskit Runtime backend.
    This allows easy access to Qiskit Runtime backends through the qrisp interface.
    It is imporant to close the session after the execution of the algorithm
    (as reported in the example below).

    Parameters
    ----------
    backend : str, optional
        A string associated to the name of a Qiskit Runtime Backend.
        The default one is the "ibmq_qasm_simulator", but it also 
        possible to provide other ones like "simulator_statevector"
        or real backends from the Qiskit Runtime.
    port : int, optional
        The port to listen. The default is 8079.
    token : str
        The token is necessary to create correctly the Qiskit Runtime
        service and be able to run algorithms on their backends.

    Example
    --------

    >>> from qrisp import QuantumFloat
    >>> from qrisp.interface import QiskitRuntimeBackend
    >>> example_backend = QiskitRuntimeBackend(backend="simulator_statevector",token='YOUR_TOKEN')
    >>> qf = QuantumFloat(4)
    >>> qf[:] = 3
    >>> res = qf*qf
    >>> result=res.get_measurement(backend = example_backend)
    >>> print(results)
    >>> example_backend.close_session()
    {9: 1.0}

    """

    def __init__(self, backend=None, port=8079, token=''):
        from qiskit_ibm_runtime import QiskitRuntimeService,Sampler,Session,Estimator

        service = QiskitRuntimeService(channel="ibm_quantum", token=token)
        if backend is None:
            backend = service.get_backend("ibmq_qasm_simulator")
        else:
            backend = service.get_backend(backend)

            
        session = Session(service,backend)
        sampler = Sampler(session=session)

        # Create the run method
        def run(qasm_str, shots, token = ""):
            # Convert to qiskit
            from qiskit import QuantumCircuit
            qiskit_qc = QuantumCircuit.from_qasm_str(qasm_str)

            from qiskit import transpile

            qiskit_qc = transpile(qiskit_qc, backend=backend)
            # Run Circuit with the Sampler Primitive
            qiskit_result = sampler.run(qiskit_qc, shots=shots).result().quasi_dists[0].binary_probabilities()
            qiskit_result.update((key, round(value * shots)) for key, value in qiskit_result.items())
            
            # Remove the spaces in the qiskit result keys
            result_dic = {}
            import re

            for key in qiskit_result.keys():
                counts_string = re.sub(r"\W", "", key)
                result_dic[counts_string] = qiskit_result[key]
            
            return result_dic

        super().__init__(run, port=port)

    def close_session(self): 
        """
        Method to call in order to close the session started by the init method.
        """
        self.session.close()


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp import x, rx

def qiro_RXMixer(solutions = [],exclusions = []):
    """
    RX-Mixer for QIRO algorithm. Works analogously to the normal RX-Mixer, but respects solutions and exclusions that have been found in the QIRO reduction steps.

    Parameters
    ----------
    solutions: List
        Solutions that have been found in the QIRO reduction steps.
    exclusions: List
        Solutions that have been found in the QIRO reduction steps.

    Returns
    -------
    RX_mixer: function
        The RX-mixer, according to the update steps that have been undertaken

    """
    union = solutions + exclusions
    def RX_mixer(qv, beta):

        for i in range(len(qv)):
            if not i in union:
                rx(2 * beta, qv[i])
    return RX_mixer


def qiro_init_function(solutions = [], exclusions = []):
    """
    State initiation function for QIRO algorithm. Works analogously to the normal initiation function, but respects solutions and exclusions that have been found in the QIRO reduction steps.

    Parameters
    ----------
    solutions: List
        Solutions that have been found in the QIRO reduction steps.
    exclusions: List
        Solutions that have been found in the QIRO reduction steps.

    Returns
    -------
    init_state: function
        The state initiation function, according to the update steps that have been undertaken

    """
    union = solutions + exclusions
    def init_state(qv):
        from qrisp import h
        #for i in problem.nodes:
        for i in range(len(qv)):
            if not i in union:
                h(qv[i])
        for i in solutions:
            x(qv[i])
    return init_state



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.arithmetic.adders.qcla.quantum_quantum.qq_carry_path import *
from qrisp.arithmetic.adders.qcla.quantum_quantum.qq_sum_path import *
from qrisp.arithmetic.adders.qcla.quantum_quantum.qq_qcla_adder import *

from qrisp.arithmetic.adders.qcla.classical_quantum.cq_carry_path import *
from qrisp.arithmetic.adders.qcla.classical_quantum.cq_sum_path import *
from qrisp.arithmetic.adders.qcla.classical_quantum.cq_qcla_adder import *


def qcla(a, b, radix_base = 2, radix_exponent = 1, t_depth_reduction = True, ctrl = None):
    r"""
    Implementation of the higher radix quantum carry lookahead adder (QCLA) as described 
    `here <https://arxiv.org/abs/2304.02921>`_. This adder stands out for having logarithmic
    T-depth like `Drapers QCLA <https://arxiv.org/abs/quant-ph/0406142>`_. Compared to Drapers
    QCLA, the higher radix QCLA allows a more dynamic structure and the use of customizable
    "sub-adders" which enables this adder to beat Drapers QCLA in terms of speed (ie. T-depth).
    
    In Python syntax, this function performs the inplace addition:
        
    ::
        
        b += a
    
    
    Apart from the two quantum arguments, this function supports the specification of the
    adder-radix R. The adder-radix can be specified in the form of an exponential of integers:
        
    .. math::
        
        R = r^k
    
    Where $r$ is the radix base and $k$ is the radix exponent.
    
    Calling ``qcla`` with radix base $r$ and radix exponent $k$, will precalculate the
    carry values using the `Brent-Kung tree <https://en.wikipedia.org/wiki/Brent%E2%80%93Kung_adder>`_
    with carry-radix $r$ and cancel the recursion $k$ layers before conclusion.
    
    An additional compilation option is given with the ``t_depth_reduction`` keyword.
    This compilation option modifies the way the carry values are uncomputed.
    If ``t_depth_reduction`` is set to ``True`` the carry values will be uncomputed using
    the intermediate result of the sub-adder - if set to ``False`` they will be uncomputed
    using the automatic uncomputation algorithm.
    
    The advantage of the automated version is that, both T-depth and CNOT-depth are scaling
    with the the logarithm of the input size. For ``t_depth_reduction = True`` the T-depth
    is significantly reduced (and still logarithmic) however the CNOT depth becomes linear.
    

    Parameters
    ----------
    a : QuantumFloat or List[Qubit] or int
        The value that is added.
    b : QuantumFloat or List[Qubit]
        The value that is operated on.
    radix_base : integer, optional
        The radix of the Brent-Kung tree. The default is 2.
    radix_exponent : integer, optional
        The cancellation treshold for the Brent-Kung recursion. The adder-Radix is then
        $R = r^k$. The default is 1.
    t_depth_reduction : bool, optional
        A compilation option that reduces T-depth but in turn weakens CNOT depth
        to linear scaling. The default is True.

    Raises
    ------
    Exception
        Tried to add QuantumFloat of higher precision onto QuantumFloat of lower precision.

    Examples
    --------
    
    We try out several constellations of parameters:
        
    >>> from qrisp import QuantumFloat, qcla
    >>> a = QuantumFloat(8)
    >>> b = QuantumFloat(8)
    >>> a[:] = 4
    >>> b[:] = 15
    >>> qcla(a, b)
    >>> print(b)
    {19: 1.0}
    
    We now measure the T-depth. To get the optimal result, we need to tell the compiler
    that we only care about T-gates. This can be achieved with the ``gate_speed`` keyword
    of the :meth:`compile <qrisp.QuantumSession.compile>` method. This keyword allows
    you to specify a function of :ref:`Operation` objects, which returns the speed
    of that Operation. For more information check out the 
    :meth:`compile <qrisp.QuantumSession.compile>` documentation page.
    
    For T-depth, there is already a pre-coded function: :ref:`qrisp.t_depth_indicator`.
        
    >>> from qrisp import t_depth_indicator
    >>> gate_speed = lambda x : t_depth_indicator(x, epsilon = 2**-10)
    >>> qc = b.qs.compile(gate_speed = gate_speed, compile_mcm = True)
    >>> qc.t_depth()
    17
    
    This function contains many allocations/deallocations that can be leveraged into
    parallelism, implying it can profit a lot from additional workspace:

    >>> qc = b.qs.compile(workspace = 10, gate_speed = gate_speed, compile_mcm = True)
    >>> qc.t_depth()        
    7
    
    We can verify the logarithmic behavior by comparing to the 
    `Gidney-adder <https://arxiv.org/abs/1709.06648>`_:
        
    >>> from qrisp import gidney_adder
    >>> a = QuantumFloat(40)
    >>> b = QuantumFloat(40)
    >>> gidney_adder(a, b)
    >>> qc = b.qs.compile(gate_speed = gate_speed, compile_mcm = True)
    >>> qc.t_depth()
    40

    >>> a = QuantumFloat(40)
    >>> b = QuantumFloat(40)
    >>> qcla(a, b)
    >>> qc = b.qs.compile(workspace = 50, gate_speed = gate_speed, compile_mcm = True)
    >>> qc.t_depth()    
    19
    
    The function can also be used to perform semi-classical in-place addition
    
    >>> b = QuantumFloat(10)
    >>> b[:] = 20
    >>> qcla(22, b)
    >>> print(b)
    {42: 1.0}
    """
    
    if isinstance(a, (int, str)):
        return cq_qcla(a, b, radix_base = radix_base, radix_exponent = radix_exponent, t_depth_reduction = t_depth_reduction, ctrl = ctrl)
    elif isinstance(a, (list, QuantumVariable)):
        return qq_qcla(a, b, radix_base = radix_base, radix_exponent = radix_exponent, t_depth_reduction = t_depth_reduction)
    else:
        raise Exception(f"Don't know how to handle type {type(a)} for QCLA addition")
    
    


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp import QuantumCircuit, QuantumVariable
from qrisp.mcx_algs.circuit_library import amy_toffoli_qc

#Implementation of the depth 2 Toffoli found in https://arxiv.org/pdf/1206.0758.pdf
def amy_toffoli(ctrl, target, ctrl_state = "11"):
    
    qc = QuantumCircuit(4)
    
    for i in [0,1]:
        if ctrl_state[i] == "0":
            qc.x(i)
    
    qc.append(amy_toffoli_qc.to_gate("amy_qc"), qc.qubits)
    
    for i in [0,1]:
        if ctrl_state[i] == "0":
            qc.x(i)
    
    amy_anc = QuantumVariable(1)
    
    amy_gate = qc.to_gate(name = "amy_toffoli")
    amy_gate.permeability = {0 : True, 1 : True, 2 : False, 3 : False}
    amy_gate.is_qfree = True
    
    amy_anc.qs.append(amy_gate, list(ctrl) + list(target) + list(amy_anc))
    amy_anc.delete()


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

def test_QAE():
    from qrisp import QuantumBool, ry, z, QAE
    import numpy as np

    def state_function(qb):
        ry(np.pi/4,qb)

    def oracle_function(qb):   
        z(qb)

    qb = QuantumBool()

    res = QAE([qb], state_function, oracle_function, precision=3)
    assert res.get_measurement() == {0.125: 0.5, 0.875: 0.5}


def test_QAE_integration():
    from qrisp import QuantumFloat, QuantumBool, control, z, h, ry, QAE
    import numpy as np

    # We compute the integral of f(x)=(sin(x))^2 from 0 to 1
    def state_function(inp, tar):
        h(inp) # Distribution
    
        N = 2**inp.size
        for k in range(inp.size):
            with control(inp[k]):
                ry(2**(k+1)/N,tar)
    
    def oracle_function(inp, tar):
        z(tar)

    n = 6 # 2^n sampling points for integration
    inp = QuantumFloat(n,-n)
    tar = QuantumBool()
    input_list = [inp, tar]

    prec = 3 # precision
    res = QAE(input_list, state_function, oracle_function, precision=prec)
    assert res.get_measurement() == {0.125: 0.31334, 0.875: 0.31334, 0.25: 0.12557, 0.75: 0.12557, 0.0: 0.05096, 0.375: 0.02632, 0.625: 0.02632, 0.5: 0.01858}


